from collections import Counter
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pickle
from datetime import datetime
from eflips.depot.layout_opt.settings import OPT_CONSTANTS as OC
from eflips.depot.evaluation import baseplot, savefig
from eflips.settings import globalConstants
from eflips.helperFunctions import cm2in


class OptEvaluation:
    """Container and tools for evaluation.

    feasible: [function] that returns the feasiblity of an individual.
    feasible_fr_vec: [function] that returns a feasibility vector of an
        individual from its results.
    """

    def __init__(
        self,
        pop,
        logbook,
        hof,
        crossovers,
        mutations,
        memory,
        feasible,
        feasible_fr_vec,
    ):
        self.pop = pop
        self.logbook = logbook
        self.hof = hof
        self.crossovers = crossovers
        self.mutations = mutations
        self.memory = memory
        self.feasible = feasible
        self.feasible_fr_vec = feasible_fr_vec

        self.OC = OC

        self.fitness_counter = None
        self.feasiblity_counter = None

        self.path_results = globalConstants["depot"]["path_results"]

    def _count_fitnesses(self):
        """Gather fitness and feasibility data."""
        if self.fitness_counter is None:
            self.fitness_counter = Counter()
            self.feasiblity_counter = Counter()

            for ID, value in self.memory.items():
                fit = self.memory[ID]["fitness"]
                count = self.memory[ID]["count"]

                self.fitness_counter[fit] += count
                fea = self.feasible_fr_vec(value["results"])
                self.feasiblity_counter[fea] += count

    def results_operators(self):
        evals = self.logbook.select("evals")
        looked_up = self.logbook.select("looked_up")
        if len(evals) == 1:
            print("No crossovers or mutations.")
            return
        print(
            "Crossovers total: %s; Average per gen: %s"
            % (sum(self.crossovers), np.average(self.crossovers))
        )
        mut_per_action = Counter(self.mutations)
        print(
            "Successful mutations: %s; Total: %s; Average per gen: %s"
            % (
                mut_per_action,
                sum(mut_per_action.values()),
                sum(mut_per_action.values()) / (len(evals) - 1),
            )
        )
        sum_evals = sum(evals)
        sum_looked_up = sum(looked_up)
        print("Total evals (incl. looked-up):", sum_evals + sum_looked_up)

    def results_simtimes(self):
        simtimes = []
        for value in self.memory.values():
            simtimes.append(value["results"]["simtime"])
        print(
            "Total simulations: %d, average computation time per simulation: %f seconds"
            % (len(simtimes), np.average(simtimes))
        )

    def results_feasbility(self):
        self._count_fitnesses()

        total_feasible = sum(
            self.feasiblity_counter[key]
            for key in self.feasiblity_counter
            if False not in key
        )
        total_infeasible = sum(self.feasiblity_counter.values()) - total_feasible
        share_feasible = total_feasible / sum(self.feasiblity_counter.values())
        print(
            "Among evals total feasible: %d, total infeasible: %d, share of feasbile: %s, Counter: %s"
            % (
                total_feasible,
                total_infeasible,
                share_feasible,
                self.feasiblity_counter,
            )
        )

    def progress_so(self, objective_label="Kapazität"):
        """Progress of the objective value for depot layout optimization with a
         single objective.

        objective_label: axis label for the objective
        """
        if len(self.pop[0].fitness.values) != 1:
            print("progress_so is for single objective only.")
            return

        (
            minimum,
            maximum,
            avg,
        ) = self.logbook.chapters[
            "fitness"
        ].select("min", "max", "avg")
        evals = self.logbook.select("evals")

        print(
            "Gen %s: min=%s, max=%s, avg=%s"
            % (len(minimum), minimum[-1], maximum[-1], avg[-1])
        )

        plt.scatter(range(len(evals)), minimum)
        plt.scatter(range(len(evals)), maximum)
        plt.scatter(range(len(evals)), avg)

        plt.grid(True)
        plt.xlabel("Generation")
        # plt.ylabel('Total capacity (absmax={})'.format(maxfitness_estimate))
        plt.ylabel(objective_label)
        plt.legend(["min", "max", "avg"])

    def progress_c_d_j(
        self,
        objectives_labels=("Kapazität", "Verspätung [h]", "Stau [h]"),
        objectives_names=("c", "d", "j"),
        show=True,
        save=False,
        formats=("pdf",),
        basefilename="progress_c_d_j",
    ):
        """Plots for the progress of the objectives values for depot layout
        optimization with objectives c, d and j. Time unit is hours.
        """
        return self.progress_mo(
            objectives_labels, objectives_names, show, save, formats, basefilename
        )

    def progress_c_urfd(
        self,
        objectives_labels=("Kapazität", "Unblocked ready for departure vehicles"),
        objectives_names=("c", "urfd"),
        show=True,
        save=False,
        formats=("pdf",),
        basefilename="progress_c_urfd",
    ):
        """Plots for the progress of the objectives values for depot layout
        optimization with objectives c and urfd.
        """
        return self.progress_mo(
            objectives_labels, objectives_names, show, save, formats, basefilename
        )

    def progress_mo(
        self,
        objectives_labels,
        objectives_names,
        show=True,
        save=False,
        formats=("pdf",),
        basefilename="progress_mo",
    ):
        """Plots for the progress of the objectives values for depot layout
        optimization with multiple objectives.

        objectives_labels: iterable with axis labels for objectives
        objectives_names: iterable with names of objectives for the filename
        """
        n_objectives = len(self.pop[0].fitness.values)
        if n_objectives < 2:
            print("progress_mo is for multiple objective only.")
            return

        (
            minimum,
            maximum,
            avg,
        ) = self.logbook.chapters[
            "fitness"
        ].select("min", "max", "avg")
        evals = self.logbook.select("evals")
        entries = len(evals)

        min_sep = []
        max_sep = []
        avg_sep = []
        for i in range(n_objectives):
            min_sep.append([fit[i] for fit in minimum])
            max_sep.append([fit[i] for fit in maximum])
            avg_sep.append([fit[i] for fit in avg])

        x = list(range(entries))

        for i in range(n_objectives):
            fig, ax = baseplot(show)
            ax.scatter(x, min_sep[i], s=12, marker=".", alpha=0.8)
            ax.scatter(x, max_sep[i], s=12, marker=".", alpha=0.8)
            ax.scatter(x, avg_sep[i], s=12, marker=".", alpha=0.8)

            plt.grid(True)
            plt.xlabel("Generation")
            plt.ylabel(objectives_labels[i])
            plt.legend(["Minimum", "Maximum", "Mittelwert"])

            dpi = fig.get_dpi()
            fig.set_size_inches(1920.0 / float(dpi), 948.0 / float(dpi))

            if show:
                fig.show()
            if save:
                filename = self.path_results + basefilename + "_" + objectives_names[i]
                savefig(fig, filename, formats)
            if not show:
                plt.close(fig)

        return min_sep, max_sep, avg_sep

    def objective_space_c_d_j(
        self,
        include_infeasible=True,
        show=True,
        save=False,
        formats=("pdf",),
        basefilename="objective_space_c_d_j",
    ):
        """Plot the objective space based on found fitnesses of depot layout
        optimization with objectives c, d, j. Fitnesses found multiple times
        are plotted only once.

        Time unit is hours.
        """
        any_value = next(iter(self.memory.values()))
        if len(any_value["fitness"]) != 3:
            print("objective_space_c_d_j requires three objectives in fitness")
            return

        x_f, y_f, z_f, x_i, y_i, z_i = ([] for _ in range(6))

        for value in self.memory.values():
            if all(self.feasible_fr_vec(value["results"])):
                x_f.append(value["fitness"][0])
                y_f.append(value["fitness"][1])
                z_f.append(value["fitness"][2])
            elif include_infeasible:
                x_i.append(value["fitness"][0])
                y_i.append(value["fitness"][1])
                z_i.append(value["fitness"][2])

        fig = plt.figure()
        ax = Axes3D(fig)

        sc_fea = ax.scatter(x_f, y_f, z_f)

        if include_infeasible:
            sc_infea = ax.scatter(x_i, y_i, z_i, c="r")
            plt.legend((sc_fea, sc_infea), ("Zulässig", "Unzulässig"))
        else:
            plt.legend((sc_fea,), ("Zulässig",))

        ax.set_xlabel("c")
        ax.set_ylabel("d [h]")
        ax.set_zlabel("j [h]")
        # plt.title('Objective Space')

        plt.grid()

        dpi = fig.get_dpi()
        fig.set_size_inches(1920.0 / float(dpi), 948.0 / float(dpi))

        if show:
            fig.show()
        if save:
            filename = self.path_results + basefilename
            savefig(fig, filename, formats)
        if not show:
            plt.close(fig)

    def pareto_front_c_d_j(
        self,
        include_infeasible=True,
        show=True,
        save=False,
        formats=("pdf",),
        basefilename="pareto_front_c_d_j",
    ):
        """Same as objective_space_c_d_j but only showing the pareto front.
        Time unit is hours.
        """
        if len(self.hof.items[0].fitness.values) != 3:
            print("pareto_front_c_d_j requires three objectives in fitness")
            return

        x_f, y_f, z_f, x_i, y_i, z_i = ([] for _ in range(6))

        for ind in self.hof.items:
            if self.feasible(ind):
                x_f.append(ind.fitness.values[0])
                y_f.append(ind.fitness.values[1])
                z_f.append(ind.fitness.values[2])
            elif include_infeasible:
                x_i.append(ind.fitness.values[0])
                y_i.append(ind.fitness.values[1])
                z_i.append(ind.fitness.values[2])

        fig = plt.figure()
        ax = Axes3D(fig)

        sc_fea = ax.scatter(x_f, y_f, z_f)

        if include_infeasible:
            sc_infea = ax.scatter(x_i, y_i, z_i, c="r")
            plt.legend((sc_fea, sc_infea), ("Zulässig", "Unzulässig"))
        else:
            plt.legend((sc_fea,), ("Zulässig",))

        ax.set_xlabel("Kapazität")
        ax.set_ylabel("Verspätung [h]")
        ax.set_zlabel("Stau [h]")
        # plt.title('Pareto Front')

        plt.grid()

        dpi = fig.get_dpi()
        fig.set_size_inches(1920.0 / float(dpi), 948.0 / float(dpi))

        if show:
            fig.show()
        if save:
            filename = self.path_results + basefilename
            savefig(fig, filename, formats)
        if not show:
            plt.close(fig)

    def objective_space_c_urfd(
        self,
        include_infeasible=True,
        show=True,
        save=False,
        formats=("pdf",),
        basefilename="objective_space_c_urfd",
        xlim=("min", "max"),
    ):
        """Plot the objective space based on found fitnesses of depot layout
        optimization with objectives c and urfd. Fitnesses found multiple times
        are plotted only once.
        """
        any_value = next(iter(self.memory.values()))
        if len(any_value["fitness"]) != 2:
            print("objective_space_c_urfd requires two objectives in fitness")
            return

        # Points due to feasibility counter capacity, delay and congestion
        (
            x_f,
            y_f,
            x_f_hof,
            y_f_hof,
            x_i_000,
            y_i_000,
            x_i_001,
            y_i_001,
            x_i_010,
            y_i_010,
            x_i_011,
            y_i_011,
            x_i_100,
            y_i_100,
            x_i_101,
            y_i_101,
            x_i_110,
            y_i_110,
        ) = ([] for _ in range(18))

        # Counter for verification
        count_f = (
            count_hof
        ) = (
            count_000
        ) = count_001 = count_010 = count_011 = count_100 = count_101 = count_110 = 0

        for value in self.memory.values():
            if all(self.feasible_fr_vec(value["results"])):
                x_f.append(value["fitness"][0])
                y_f.append(value["fitness"][1])
                count_f += 1
            elif include_infeasible:
                if not value["results"]["feasible_capacity"]:
                    if not value["results"]["feasible_delay"]:
                        if not value["results"]["feasible_congestion"]:
                            x_i_000.append(value["fitness"][0])
                            y_i_000.append(value["fitness"][1])
                            count_000 += 1
                        else:
                            x_i_001.append(value["fitness"][0])
                            y_i_001.append(value["fitness"][1])
                            count_001 += 1
                    else:
                        if not value["results"]["feasible_congestion"]:
                            x_i_010.append(value["fitness"][0])
                            y_i_010.append(value["fitness"][1])
                            count_010 += 1
                        else:
                            x_i_011.append(value["fitness"][0])
                            y_i_011.append(value["fitness"][1])
                            count_011 += 1
                else:
                    if not value["results"]["feasible_delay"]:
                        if not value["results"]["feasible_congestion"]:
                            x_i_100.append(value["fitness"][0])
                            y_i_100.append(value["fitness"][1])
                            count_100 += 1
                        else:
                            x_i_101.append(value["fitness"][0])
                            y_i_101.append(value["fitness"][1])
                            count_101 += 1
                    else:
                        if not value["results"]["feasible_congestion"]:
                            x_i_110.append(value["fitness"][0])
                            y_i_110.append(value["fitness"][1])
                            count_110 += 1
                        else:
                            print("All feasible option already plotted")

        for ind in self.hof.items:
            if self.feasible(ind):
                x_f_hof.append(ind.fitness.values[0])
                y_f_hof.append(ind.fitness.values[1])
                count_hof += 1

        print(
            count_f,
            count_hof,
            count_000,
            count_001,
            count_010,
            count_011,
            count_100,
            count_101,
            count_110,
        )

        fig, ax = baseplot(show)

        if include_infeasible:
            sc_infea_000 = ax.scatter(
                x_i_000, y_i_000, marker="x", c="#969696", alpha=0.3
            )
            sc_infea_001 = ax.scatter(
                x_i_001, y_i_001, marker="x", c="#000096", alpha=0.3
            )
            sc_infea_010 = ax.scatter(
                x_i_010, y_i_010, marker="x", c="#009600", alpha=0.3
            )
            sc_infea_011 = ax.scatter(
                x_i_011, y_i_011, marker="x", c="#009696", alpha=0.3
            )
            sc_infea_100 = ax.scatter(
                x_i_100, y_i_100, marker="x", c="#960000", alpha=0.3
            )
            sc_infea_101 = ax.scatter(
                x_i_101, y_i_101, marker="x", c="#960096", alpha=0.3
            )
            sc_infea_110 = ax.scatter(
                x_i_110, y_i_110, marker="x", c="#969600", alpha=0.3
            )
            sc_fea = ax.scatter(x_f, y_f, marker=".", c="k", alpha=0.3)
            sc_fea_hof = ax.scatter(x_f_hof, y_f_hof, marker=".", c="orange")
            plt.legend(
                (
                    sc_fea_hof,
                    sc_fea,
                    sc_infea_000,
                    sc_infea_001,
                    sc_infea_010,
                    sc_infea_011,
                    sc_infea_100,
                    sc_infea_101,
                    sc_infea_110,
                ),
                (
                    "Pareto-Menge",
                    "Zulässig",
                    "Unzulässig 000",
                    "Unzulässig 001",
                    "Unzulässig 010",
                    "Unzulässig 011",
                    "Unzulässig 100",
                    "Unzulässig 101",
                    "Unzulässig 110",
                ),
                loc="upper left",
            )
        else:
            sc_fea = ax.scatter(x_f, y_f, marker=".", c="k", alpha=0.3)
            sc_fea_hof = ax.scatter(x_f_hof, y_f_hof, marker=".", c="orange")
            plt.legend(
                (sc_fea_hof, sc_fea), ("Pareto-Menge", "Zulässig"), loc="upper left"
            )

        ax.set_xlabel("c")
        ax.set_ylabel(r"$\bar{n}_{v,urfd}$")
        # plt.title('Objective Space')

        plt.grid()
        left, right = plt.xlim()
        xlim = list(xlim)
        if xlim[0] == "min":
            xlim[0] = left
        if xlim[1] == "max":
            xlim[1] = right
        plt.xlim(*xlim)

        dpi = fig.get_dpi()
        # fig.set_size_inches(1920.0 / float(dpi), 948.0 / float(dpi))
        fig.set_size_inches(19 / 2.54, 10 / 2.54)

        if show:
            fig.show()
        if save:
            filename = self.path_results + basefilename
            savefig(fig, filename, formats)
        if not show:
            plt.close(fig)

    def pareto_front_c_urfd(
        self,
        include_infeasible=True,
        show=True,
        save=False,
        formats=("pdf",),
        basefilename="pareto_front_c_urfd",
    ):
        """Same as objective_space_c_urfd but only showing the pareto front."""
        if len(self.hof.items[0].fitness.values) != 2:
            print("pareto_front_c_urfd requires two objectives in fitness")
            return

        x_f, y_f, x_i, y_i = ([] for _ in range(4))

        for ind in self.hof.items:
            if self.feasible(ind):
                x_f.append(ind.fitness.values[0])
                y_f.append(ind.fitness.values[1])
            elif include_infeasible:
                x_i.append(ind.fitness.values[0])
                y_i.append(ind.fitness.values[1])

        fig, ax = baseplot(show)
        sc_fea = ax.scatter(x_f, y_f)

        if include_infeasible:
            sc_infea = ax.scatter(x_i, y_i, c="r")
            plt.legend((sc_fea, sc_infea), ("Zulässig", "Unzulässig"))
        else:
            plt.legend((sc_fea,), ("Zulässig",))

        ax.set_xlabel("c")
        ax.set_ylabel("urfd")
        # plt.title('Pareto Front')

        plt.grid()

        dpi = fig.get_dpi()
        # fig.set_size_inches(1920.0 / float(dpi), 948.0 / float(dpi))
        fig.set_size_inches(13 / 2.54, 5 / 2.54)

        if show:
            fig.show()
        if save:
            filename = self.path_results + basefilename
            savefig(fig, filename, formats)
        if not show:
            plt.close(fig)

    def comptime(
        self,
        show=True,
        save=False,
        formats=("pdf",),
        basefilename="comptime",
        xlim=(0, "max"),
    ):
        """Computation time and number of lookups per generation. When
        multiprocessing was used, values still represent wall clock time, not
        computation time sum of worker processes.
        """
        ct, lu, sk = self.logbook.select("comptime", "looked_up", "skipped")
        entries = len(ct)
        ngen = entries - 1
        x = range(entries)

        xlim = list(xlim)
        if xlim[1] == "max":
            xlim[1] = ngen

        N = max(int(xlim[1] * 0.08), 1)  # precision of running mean
        ct_trend = running_mean(ct, N)
        lu_trend = running_mean(lu, N)
        sk_trend = running_mean(sk, N)

        fig, ax1 = baseplot(show)
        plt.xlabel("Generation")

        # c_ct = '#1f77b4'  # blue
        c_ct = "#d62728"  # red
        ax1.plot(ct_trend, c=c_ct)
        ct_plot = ax1.scatter(x, ct, s=12, marker=".", alpha=0.5, c=c_ct)
        ax1.set_ylabel("Sekunden", c=c_ct)

        ax2 = ax1.twinx()

        # ax2.plot(lu_trend, c='#ff7f0e')
        # lu_plot = ax2.scatter(x, lu, s=12, marker='.', alpha=0.5, c='#ff7f0e')
        # # ax2.set_ylabel('Looked up', c='#ff7f0e')
        ax2.set_ylabel("Anzahl", c="#2ca02c")

        ax2.plot(sk_trend, c="#2ca02c")
        sk_plot = ax2.scatter(x, sk, s=12, marker=".", alpha=0.5, c="#2ca02c")
        # ax2.set_ylabel('Skipped', c='#2ca02c')

        ax2.set_ylim(bottom=0, top=len(self.pop))

        # plt.legend((ct_plot, lu_plot, sk_plot), ('Rechenzeit [s]', 'Bekannt', 'Keine Simulation wg. vermuteter Unzulässigkeit'))
        plt.legend((ct_plot, sk_plot), ("Rechenzeit", "Vermutl. unzulässig"))

        plt.xlim(*xlim)
        ax1.grid()

        # dpi = fig.get_dpi()
        # fig.set_size_inches(1920.0 / float(dpi), 948.0 / float(dpi))
        fig.set_size_inches(cm2in(16), cm2in(8))

        if show:
            fig.show()
        if save:
            filename = self.path_results + basefilename
            savefig(fig, filename, formats)
        if not show:
            plt.close(fig)

    def ideas(
        self,
        show=True,
        save=False,
        formats=("pdf",),
        basefilename="ideas",
        xlim=(0, "max"),
    ):
        """Number of individuals with new and looked up fitness by generation."""
        evals, lu = self.logbook.select("evals", "looked_up")
        entries = len(evals)
        ngen = entries - 1
        x = range(entries)

        xlim = list(xlim)
        if xlim[1] == "max":
            xlim[1] = ngen

        N = max(int(xlim[1] * 0.08), 1)  # precision of running mean
        evals_trend = running_mean(evals, N)
        lu_trend = running_mean(lu, N)

        fig, ax = baseplot(show)
        plt.xlabel("Generation")

        ax.plot(evals_trend)
        evals_plot = ax.scatter(x, evals, s=12, marker=".", alpha=0.5)

        ax.plot(lu_trend)
        lu_plot = ax.scatter(x, lu, s=12, marker=".", alpha=0.5)

        ax.set_ylabel("Anzahl")

        plt.legend((evals_plot, lu_plot), ("Neu", "Bekannt"), loc="upper right")

        plt.grid()
        xlim = list(xlim)
        if xlim[1] == "max":
            xlim[1] = ngen
        plt.xlim(*xlim)

        # dpi = fig.get_dpi()
        # fig.set_size_inches(1920.0 / float(dpi), 948.0 / float(dpi))
        fig.set_size_inches(cm2in(16), cm2in(8))

        plt.ylim(0, len(self.pop))

        if show:
            fig.show()
        if save:
            filename = self.path_results + basefilename
            savefig(fig, filename, formats)
        if not show:
            plt.close(fig)

    def save(self, filename=None):
        if filename is None:
            filename = self.path_results + "results_" + now_repr()

        filename += ".p"
        with open(filename, "wb") as file:
            pickle.dump(self, file)
        print("Saved as %s" % filename)

    def draw_hof(self, save=False, formats=("pdf",), basefilename="hof"):
        for ind in self.hof:
            if self.feasible(ind):
                print(ind.ID, ind.fitness.values)
                ind.visu.repack()
                fig, ax = ind.visu.draw()
                if save:
                    filename = self.path_results + basefilename + "_" + ind.ID
                    savefig(fig, filename, formats)


def now_repr():
    """Return the current system date and time as formatted string."""
    return datetime.now().strftime("%Y-%m-%d_%H-%M-%S")


def load(filename):
    """Load from pickle file.
    When loading OptEvaluation, the setup in the optimizing execution file must
    be run first so that classes such as Individual exist.

    filename: [str] including path, excluding extension
    """
    filename += ".p"
    with open(filename, "rb") as file:
        obj = pickle.load(file)
    return obj


def running_mean(x, N):
    """Calculate the running mean of *x*.
    From https://stackoverflow.com/a/27681394
    """
    cumsum = np.cumsum(np.insert(x, 0, 0))
    return (cumsum[N:] - cumsum[:-N]) / float(N)



"""This package contains the private API for the depot-related functionality in eFLIPS."""

from datetime import timedelta
from enum import Enum, auto
from math import ceil
from typing import Dict, List, Tuple

import sqlalchemy.orm
from eflips.model import (
    Scenario,
    AssocPlanProcess,
    Area,
    AssocAreaProcess,
    Event,
    Process,
    Depot,
    Plan,
    AreaType,
    Rotation,
    Trip,
    Station,
    VehicleType,
)
from sqlalchemy.orm import Session


def delete_depots(scenario: Scenario, session: Session) -> None:
    """This function deletes all depot-related data from the database for a given scenario.

    Used before a new depot in this scenario is created.

    :param scenario: The scenario to be simulated
    :param session: The database session

    :return: None. The depot-related data will be deleted from the database.
    """

    # Delete assocs
    session.query(AssocPlanProcess).filter(
        AssocPlanProcess.scenario_id == scenario.id
    ).delete()
    list_of_area = session.query(Area).filter(Area.scenario_id == scenario.id).all()

    for area in list_of_area:
        session.query(AssocAreaProcess).filter(
            AssocAreaProcess.area_id == area.id
        ).delete()
        session.query(Event).filter(Event.area_id == area.id).delete()

    # delete processes
    session.query(Process).filter(Process.scenario_id == scenario.id).delete()

    # delete areas
    session.query(Area).filter(Area.scenario_id == scenario.id).delete()
    # delete depot
    session.query(Depot).filter(Depot.scenario_id == scenario.id).delete()
    # delete plan
    session.query(Plan).filter(Plan.scenario_id == scenario.id).delete()
    # delete assoc_plan_process


def depot_to_template(depot: Depot) -> Dict[str, str | Dict[str, str | int]]:
    """
    Converts the depot to a template for internal use in the simulation core.

    :return: A dict that can be consumed by eFLIPS-Depot.
    """
    # Initialize the template
    template = {
        "templatename_display": "",
        "general": {"depotID": "", "dispatch_strategy_name": ""},
        "resources": {},
        "resource_switches": {},
        "processes": {},
        "areas": {},
        "groups": {},
        "plans": {},
    }

    # Set up the general information
    template["templatename_display"] = depot.name
    template["general"]["depotID"] = str(depot.id)
    template["general"]["dispatch_strategy_name"] = "SMART"

    # Helper for adding processes to the template
    list_of_processes = []

    # Get dictionary of each area
    for area in depot.areas:
        area_name = str(area.id)
        template["areas"][area_name] = {
            "typename": (
                "LineArea" if area.area_type == AreaType.LINE else "DirectArea"
            ),
            "capacity": area.capacity,
            "available_processes": [str(process.id) for process in area.processes],
            "issink": False,
            "entry_filter": None,
        }

        # Fill in vehicle_filter.
        template["areas"][area_name]["entry_filter"] = {
            "filter_names": ["vehicle_type"],
            "vehicle_types": [str(area.vehicle_type_id)],
        }

        # TODO for cleaning area etc., enable non-vehicle_type areas

        for process in area.processes:
            # Add process into process list
            list_of_processes.append(
                process
            ) if process not in list_of_processes else None

            # Charging interfaces
            if process_type(process) == ProcessType.CHARGING:
                ci_per_area = []
                for i in range(area.capacity):
                    ID = "ci_" + str(len(template["resources"]))
                    template["resources"][ID] = {
                        "typename": "DepotChargingInterface",
                        "max_power": process.electric_power,
                    }
                    ci_per_area.append(ID)

                template["areas"][area_name]["charging_interfaces"] = ci_per_area

            # Set issink to True for departure areas
            if process_type(process) == ProcessType.STANDBY_DEPARTURE:
                template["areas"][area_name]["issink"] = True

    for process in list_of_processes:
        process_name = str(process.id)
        # Shared template for all processes
        template["processes"][process_name] = {
            "typename": "",  # Placeholder for initialization
            "dur": int(process.duration.total_seconds()) if process.duration else None,
            # True if this process will be executed for all vehicles. False if there are available vehicle filters
            "ismandatory": True,
            "vehicle_filter": {},
            # True if this process can be interrupted by a dispatch. False if it cannot be interrupted
            "cancellable_for_dispatch": process.dispatchable,
        }

        match process_type(process):
            case ProcessType.SERVICE:
                template["processes"][process_name]["typename"] = "Serve"

                # Fill in the worker_service
                service_capacity = sum([x.capacity for x in process.areas])

                template["processes"][process_name]["required_resources"] = [
                    "workers_service"
                ]
                template["resources"]["workers_service"] = {
                    "typename": "DepotResource",
                    "capacity": service_capacity,
                }

                if process.availability is not None and len(process.availability) > 0:
                    template["resource_switches"]["service_switch"] = {
                        "resource": "workers_service",
                        "breaks": [],
                        "preempt": process.preemptable
                        if process.preemptable is not None
                        else True,
                        # Strength 'full' means all workers can take a break at the same time
                        "strength": "full",
                        # Resume set to True means that the process will continue after the break
                        "resume": True,
                        # Priority -3 means this process has the highest priority
                        "priority": -3,
                    }

                    list_of_breaks = process._generate_break_intervals()
                    list_of_breaks_in_seconds = []

                    # Converting the time intervals into seconds

                    for time_interval in list_of_breaks:
                        start_time = time_interval[0]
                        end_time = time_interval[1]
                        start_time_in_seconds = (
                            start_time.hour * 3600
                            + start_time.minute * 60
                            + start_time.second
                        )

                        end_time_in_seconds = (
                            end_time.hour * 3600
                            + end_time.minute * 60
                            + end_time.second
                        )

                        list_of_breaks_in_seconds.append(
                            (start_time_in_seconds, end_time_in_seconds)
                        )

                    template["resource_switches"]["service_switch"][
                        "breaks"
                    ] = list_of_breaks_in_seconds

            case ProcessType.CHARGING:
                template["processes"][process_name]["typename"] = "Charge"
                del template["processes"][process_name]["dur"]

            case ProcessType.STANDBY | ProcessType.STANDBY_DEPARTURE:
                template["processes"][process_name]["typename"] = "Standby"
                template["processes"][process_name]["dur"] = 0

            case ProcessType.PRECONDITION:
                template["processes"][process_name]["typename"] = "Precondition"
                template["processes"][process_name]["dur"] = int(
                    process.duration.total_seconds()
                )
                template["processes"][process_name]["power"] = process.electric_power
            case _:
                raise ValueError(f"Invalid process type: {process_type(process).name}")

    # Initialize the default plan
    template["plans"]["default"] = {
        "typename": "DefaultActivityPlan",
        "locations": [],
    }
    # Groups
    for process in depot.default_plan.processes:
        group_name = str(process.name) + "_group"
        template["groups"][group_name] = {
            "typename": "AreaGroup",
            "stores": [str(area.id) for area in process.areas],
        }
        if process_type(process) == ProcessType.STANDBY_DEPARTURE:
            template["groups"][group_name]["typename"] = "ParkingAreaGroup"
            template["groups"][group_name]["parking_strategy_name"] = "SMART2"

        # Fill in locations of the plan
        template["plans"]["default"]["locations"].append(group_name)

    return template


def find_first_last_stop_for_rotation_id(
    rotation: Rotation, session: sqlalchemy.orm.session.Session
) -> Tuple[Station, Station, VehicleType]:
    """
    Identifies the first stop, last stop and vehicle type for a given rotation.

    :param rotation: An :class:`eflips.model.Rotation` object
    :param session: An SQLAlchemy session object to the database
    :return: A tuple of the first stop, last stop and vehicle type
    """

    first_stop = rotation.trips[0].route.departure_station
    last_stop = rotation.trips[-1].route.arrival_station
    vehicle_type = rotation.vehicle_type
    return first_stop, last_stop, vehicle_type


def group_rotations_by_start_end_stop(
    scenario_id: int,
    session: sqlalchemy.orm.session.Session,
) -> Dict[Tuple[Station, Station], Dict[VehicleType, List[Rotation]]]:
    """
    For a given scenario, create a list of rotations and group them by their start and end stops.

    :param session: An SQLAlchemy session object
    :return: A dictionary of rotations grouped by their start and end stops, with each group further grouped by vehicle
        type.
    """
    rotations = (
        session.query(Rotation)
        .filter(Rotation.scenario_id == scenario_id)
        .options(sqlalchemy.orm.joinedload(Rotation.trips).joinedload(Trip.route))
        .options(sqlalchemy.orm.joinedload(Rotation.vehicle_type))
    )
    grouped_rotations: Dict[
        Tuple[Station, Station], Dict[VehicleType, List[Rotation]]
    ] = {}
    for rotation in rotations:
        first_stop, last_stop, vehicle_type = find_first_last_stop_for_rotation_id(
            rotation, session
        )
        if (first_stop, last_stop) not in grouped_rotations:
            grouped_rotations[(first_stop, last_stop)] = {}
        if vehicle_type not in grouped_rotations[(first_stop, last_stop)]:
            grouped_rotations[(first_stop, last_stop)][vehicle_type] = []
        grouped_rotations[(first_stop, last_stop)][vehicle_type].append(rotation)

    return grouped_rotations


def create_simple_depot(
    scenario: Scenario,
    station: Station,
    charging_capacities: Dict[VehicleType, int],
    cleaning_capacities: Dict[VehicleType, int],
    charging_power: float,
    session: sqlalchemy.orm.session.Session,
    cleaning_duration: timedelta = timedelta(minutes=30),
    safety_margin: float = 0.0,
) -> None:
    """
    Creates a simple depot for a given scenario.

    It has one area for each vehicle type and a charging process for each
    area. Also an arrival area for each vehicle type.

    :param safety_margin: a safety margin for the number of charging and cleaning capacities. Default is 0.0
    :param scenario: The scenario to be simulated
    :param station: The station where the depot is located
    :param charging_capacities: A dictionary of vehicle types and the number of vehicles that can be charged at the same time
    :param cleaning_capacities: A dictionary of vehicle types and the number of vehicles that can be cleaned at the same time
    :param charging_power: The power of the charging process
    :param cleaning_duration: The duration of the cleaning process
    :param session: An SQLAlchemy session object to the database
    :return: Nothing. Depots are created in the database.
    """

    # Create a simple depot
    depot = Depot(
        scenario=scenario,
        name=f"Depot at {station.name}",
        name_short=station.name_short,
        station=station,
    )
    session.add(depot)

    # Create plan
    plan = Plan(scenario=scenario, name=f"Default Plan")
    session.add(plan)

    depot.default_plan = plan

    # Create processes
    shunting_1 = Process(
        name="Shunting 1",
        scenario=scenario,
        dispatchable=False,
        duration=timedelta(minutes=5),
    )
    clean = Process(
        name="Arrival Cleaning",
        scenario=scenario,
        dispatchable=False,
        duration=cleaning_duration,
    )
    shunting_2 = Process(
        name="Shunting 2",
        scenario=scenario,
        dispatchable=False,
        duration=timedelta(minutes=5),
    )
    charging = Process(
        name="Charging",
        scenario=scenario,
        dispatchable=True,
        electric_power=charging_power,
    )
    standby_departure = Process(
        name="Standby Pre-departure",
        scenario=scenario,
        dispatchable=True,
    )

    session.add(clean)
    session.add(shunting_1)
    session.add(charging)
    session.add(standby_departure)
    session.add(shunting_2)

    # Create shared waiting area
    waiting_area = Area(
        scenario=scenario,
        name=f"Waiting Area for every type of vehicle",
        depot=depot,
        area_type=AreaType.DIRECT_ONESIDE,
        capacity=100,
    )
    session.add(waiting_area)

    for vehicle_type in charging_capacities.keys():
        charging_count = charging_capacities[vehicle_type]

        charging_count = int(ceil(charging_count * (1 + safety_margin)))

        # Create charging area
        charging_area = Area(
            scenario=scenario,
            name=f"Direct Charging Area for {vehicle_type.name_short}",
            depot=depot,
            area_type=AreaType.DIRECT_ONESIDE,
            capacity=int(charging_count * 1),
        )
        session.add(charging_area)
        charging_area.vehicle_type = vehicle_type

        # Create cleaning area
        cleaning_count = cleaning_capacities[vehicle_type]

        cleaning_count = int(ceil(cleaning_count * (1 + safety_margin)))

        cleaning_area = Area(
            scenario=scenario,
            name=f"Cleaning Area for {vehicle_type.name_short}",
            depot=depot,
            area_type=AreaType.DIRECT_ONESIDE,
            capacity=cleaning_count,
        )
        session.add(cleaning_area)
        cleaning_area.vehicle_type = vehicle_type

        shunting_area_1 = Area(
            scenario=scenario,
            name=f"Shunting Area 1 for {vehicle_type.name_short}",
            depot=depot,
            area_type=AreaType.DIRECT_ONESIDE,
            capacity=10,
        )

        session.add(shunting_area_1)
        shunting_area_1.vehicle_type = vehicle_type

        shunting_area_2 = Area(
            scenario=scenario,
            name=f"Shunting Area 2 for {vehicle_type.name_short}",
            depot=depot,
            area_type=AreaType.DIRECT_ONESIDE,
            capacity=10,
        )

        session.add(shunting_area_2)
        shunting_area_2.vehicle_type = vehicle_type

        cleaning_area.processes.append(clean)
        charging_area.processes.append(charging)
        charging_area.processes.append(standby_departure)
        shunting_area_1.processes.append(shunting_1)
        shunting_area_2.processes.append(shunting_2)

        assocs = [
            AssocPlanProcess(
                scenario=scenario, process=shunting_1, plan=plan, ordinal=0
            ),
            AssocPlanProcess(scenario=scenario, process=clean, plan=plan, ordinal=1),
            AssocPlanProcess(
                scenario=scenario, process=shunting_2, plan=plan, ordinal=2
            ),
            AssocPlanProcess(scenario=scenario, process=charging, plan=plan, ordinal=3),
            AssocPlanProcess(
                scenario=scenario, process=standby_departure, plan=plan, ordinal=4
            ),
        ]
        session.add_all(assocs)


class ProcessType(Enum):
    """This class represents the types of a process in eFLIPS-Depot."""

    SERVICE = auto()
    """This process represents a bus service by workers.

    It does not require a charging_power and has a fixed
    duration.
    """
    CHARGING = auto()
    """This process represents a bus charging process.

    It requires a charging_power and has no fixed duration.
    """
    STANDBY = auto()
    """This process represents an arriving bus that is waiting for a service. It does not require a charging_power.

    and has no fixed duration.
    """
    STANDBY_DEPARTURE = auto()
    """This process represents a bus ready for departure.

    It does not require a charging_power and has no fixed
    duration.
    """
    PRECONDITION = auto()
    """This process represents a bus preconditioning process.

    It requires a charging_power and has a fixed duration.
    """


def process_type(p: Process) -> ProcessType:
    """
    The type of the process.

    See :class:`eflips.depot.api.input.ProcessType` for more information. Note that whether
    a process needs a resource or not depends on the type of the process.
    """
    if p.duration is not None and p.electric_power is None:
        return ProcessType.SERVICE
    elif p.duration is None and p.electric_power is not None:
        return ProcessType.CHARGING
    elif p.duration is not None and p.electric_power is not None:
        return ProcessType.PRECONDITION
    elif p.duration is None and p.electric_power is None:
        if p.dispatchable:
            return ProcessType.STANDBY_DEPARTURE
        else:
            return ProcessType.STANDBY
    else:
        raise ValueError("Invalid process type")



"""Determination of fitness and feasibility of individuals specific for an
optimization scenario with objectives c and urfd.
"""
import eflips
from eflips.depot.layout_opt import opt_tools
from eflips.depot.layout_opt.settings import OPT_CONSTANTS as OC


PENALTY_DELTA = 0


# Load eflips.globalConstants once for all simulation runs
filename_eflips_settings = OC["scenario"]["filename_eflips_settings"]
eflips.load_settings(filename_eflips_settings)
eflips.check_gc_validity()
eflips.complete_gc()
GC = eflips.globalConstants
SIM_TIME = GC["general"]["SIMULATION_TIME"]

# Load data from excel to init a Timetable once for all simulation runs
filename_timetable = OC["scenario"]["filename_timetable"]
timetabledata = eflips.depot.standalone.timetabledata_from_excel(filename_timetable)


def estimate_max_total_delay(ttd):
    """Return an estimate of the maximum total delay that is possible until
    sim time end. SIM_TIME should be a multiple of 86400 (defining full days).

    ttd: [eflips.depot.standalone.ExcelSheetData]
    """
    n_vehicles = 0
    for depot_ID in GC["depot"]["vehicle_count"]:
        for vt_ID in GC["depot"]["vehicle_count"][depot_ID]:
            n_vehicles += GC["depot"]["vehicle_count"][depot_ID][vt_ID]

    index_std = ttd.map_headers()["std [s]"]
    count_skipped = 0
    count = 0
    estimate = 0

    for i in range(SIM_TIME // 86400):
        for tripdata in ttd.data[1:]:  # Skip header row
            if count_skipped > n_vehicles:
                std = tripdata[index_std] * (i + 1)
                estimate += SIM_TIME - std
                count += 1
            else:
                # Skip the first n_vehicles trips because they will be served
                # by vehicles from the init store and therefore cannot be
                # delayed
                count_skipped += 1

    # print('estimate_max_total_delay: %d, max std: %d, count: %d, count_skipped: %d'
    #       % (estimate, std, count, count_skipped))
    return estimate


def estimate_max_total_congestion(ttd):
    """Return an estimate of the maximum total congestion time that is possible
    until sim time end. SIM_TIME should be a multiple of 86400 (defining full
    days).

    ttd: [eflips.depot.standalone.ExcelSheetData]
    """
    index_sta = ttd.map_headers()["sta [s]"]
    count = 0
    estimate = 0

    for i in range(SIM_TIME // 86400):
        for tripdata in ttd.data[1:]:  # Skip header row
            sta = tripdata[index_sta] * (i + 1)
            if sta < SIM_TIME:  # Skip trips with sta after sim time end
                estimate += SIM_TIME - sta
                count += 1

    # print('estimate_max_total_congestion: %d, max sta: %d, count: %d'
    #       % (estimate, sta, count))
    return estimate


# Precompute data for normalizing evaluation results
for key in ["max_capacity_estimate", "max_delay_estimate", "max_congestion_estimate"]:
    if key in OC["scenario"]:
        raise ValueError(
            "Attempted to set key '%s' in OC['scenario'] but it's already used." % key
        )
OC["scenario"]["max_capacity_estimate"] = opt_tools.init.CAPACITY_MAX
OC["scenario"]["max_delay_estimate"] = estimate_max_total_delay(timetabledata)
OC["scenario"]["max_congestion_estimate"] = estimate_max_total_congestion(timetabledata)


def evaluate(ind):
    """Return the fitness tuple capacity, urfd for an individual.
    Caution: Returns other results in addition to fitness to not lose them with
    multiprocessing.
    """
    c = evaluate_capacity(ind)
    urfd = evaluate_urfd(ind)

    evaluate_delay(ind)
    evaluate_congestion(ind)

    if (
        not OC["scenario"]["simulate_below_capacity_min"]
        and ind.capacity < OC["scenario"]["CAPACITY_MIN"]
    ):
        # If the minimum capacity is violated, then skip simulation and assign
        # standard values to delay and congestion (see designated evaluate-
        # functions), to speed up evaluation. CAPACITY_MIN needs to be adapted
        # to scenarios with new depot limits!
        ind.results["simtime"] = 0

    return c, urfd, ind.results


def evaluate_single(ind):
    """Return the capacity fitness as tuple (for single objective)."""
    c = evaluate_capacity(ind)
    return (c,)


def evaluate_capacity(ind):
    """Return the objective function value for capacity."""
    ind.results["capacity"] = ind.capacity
    ind.results["feasible_capacity"] = feasible_capacity(ind)
    if ind.results["feasible_capacity"]:
        c = ind.capacity
    else:
        c = PENALTY_DELTA - distance_capacity(ind)
    return c


def evaluate_urfd(ind):
    """Objective function value for number of unblocked rfd vehicles."""
    if (
        not OC["scenario"]["simulate_below_capacity_min"]
        and ind.capacity < OC["scenario"]["CAPACITY_MIN"]
    ):
        urfd = 0
    else:
        evaluate_simulation(ind)
        urfd = ind.results["rfd_unblocked"]
    return urfd


def evaluate_delay(ind):
    """Evaluate the delay (result in hours!)."""
    if (
        not OC["scenario"]["simulate_below_capacity_min"]
        and ind.capacity < OC["scenario"]["CAPACITY_MIN"]
    ):
        ind.results["delay"] = OC["scenario"]["max_delay_estimate"] / 3600
    else:
        evaluate_simulation(ind)
    ind.results["feasible_delay"] = feasible_delay(ind)


def evaluate_congestion(ind):
    """Evaluate the congestion (result in hours!)."""
    if (
        not OC["scenario"]["simulate_below_capacity_min"]
        and ind.capacity < OC["scenario"]["CAPACITY_MIN"]
    ):
        ind.results["congestion"] = OC["scenario"]["max_congestion_estimate"] / 3600
    else:
        evaluate_simulation(ind)
    ind.results["feasible_congestion"] = feasible_congestion(ind)


def feasible_capacity(ind):
    """Do the packing test and return its result."""
    if not ind.areas:
        # an empty depot is considered infeasible
        return False

    if ind.visu.feasible is None:
        ind.visu.pack()

    return ind.visu.feasible


def feasible_delay(ind):
    """Return True if *ind* is feasible in terms of delay, i.e. there was none."""
    return ind.results["delay"] == 0


def feasible_congestion(ind):
    """Return True if *ind* is feasible in terms of congestion, i.e. there was
    none.
    """
    return ind.results["congestion"] == 0


def distance_capacity(ind):
    """Return an estimate of how far away *ind* is from the feasible region in
    terms of capacity.
    """
    # capacity always 0 if infeasible
    distance = 0

    # If the capacity is lower than the estimated max, but still not feasbile,
    # then return 0
    # distance = max(ind.capacity - OC['scenario']['max_capacity_estimate'], 0)
    return distance


def evaluate_simulation(ind):
    """Simulate and evaluate the results.

    Delay and congestion values are converted to hours.
    """
    if not ind.results["simulated"]:
        simulation_host = opt_tools.simulate(ind, timetabledata)
        ev = simulation_host.depots[0].evaluation

        # Total delay sum
        ev.total_delay()
        ind.results["delay"] = ev.results["total_delay"] / 3600

        # Total congestion sum
        ev.total_parking_congestion()
        ind.results["congestion"] = ev.results["total_parking_congestion"] / 3600

        # Unblocked rfd vehicles
        ev.calc_count_rfd_unblocked_total()
        ind.results["rfd_unblocked"] = ev.results["count_rfd_unblocked_total"]["mean"]
    else:
        # No need to simulate again
        pass


def feasible(ind):
    """Return True if *ind* is feasible. *ind* must be evaluated first."""
    return feasible_fr(ind.results)


def feasible_fr(results):
    """Return True if *ind* is feasible. *results* is an attribute of *ind*.
    *ind* must be evaluated first."""
    result = all(
        [
            results["feasible_capacity"],
            results["feasible_delay"],
            results["feasible_congestion"],
        ]
    )
    return result


def feasible_fr_vec(results):
    """Return a feasibility vector based on the evaluation results of an
    individual.
    """
    feasibility = [
        results["feasible_capacity"],
        results["feasible_delay"],
        results["feasible_congestion"],
    ]
    return tuple(feasibility)



# -*- coding: utf-8 -*-
"""
Created on Fri Oct 13 11:16:58 2017

@author: P.Mundt

Vehicle components for the depot simulation.

"""
from eflips.helperFunctions import flexprint
from eflips.settings import globalConstants
from eflips.evaluation import DataLogger
from eflips.depot.depot import DepotWorkingData
from eflips.depot.evaluation import BatteryLog


class VehicleType:
    """Definition of a vehicle type. Instantiated once per type. Vehicles of
    the same type hold references to the same VehicleType instance.

    Parameters:
    ID: [str]
    battery_capacity: [int or float] energy_nominal of battery.
        See SimpleBattery for info about other battery attributes.
    CR: [int or float] consumption rate of energy from the battery. kWh/km if
        distance-based or kW if time-based. eflips.settings.globalConstants[
        'depot']['consumption_calc_mode'] must be set accordingly. May be
        omitted if consumption calculation is not CR-based.

    Attributes:
    group: [VehicleTypeGroup or None] that the vehicle type is part of, if any.
    count: {dict} key: depot, value: number of vehicles with this type by
        depot
    share: {dict} key: depot, value: share of vehicles with this type by
        depot

    """

    def __init__(self, ID, battery_capacity, soc_min, soc_max, soc_init, soh, CR=None):
        self.ID = ID
        self.battery_capacity = battery_capacity
        self.soc_min = soc_min
        self.soc_max = soc_max
        self.soc_init = soc_init
        self.soh = soh
        self.CR = CR
        self.group = None

        self.count = {}
        self.share = {}

    def __repr__(self):
        return "{%s} %s" % (type(self).__name__, self.ID)


class VehicleTypeGroup:
    """Hashable identifier for a list of vehicle types that are mutually
    substitutable on trips.

    Parameters:
    types: [list] of VehicleType objects. Can be empty upon instantiation.

    Attributes:
    count: {dict} key: depot, value: number of vehicles with this group by
        depot
    share: {dict} key: depot, value: share of vehicles with this group by
        depot

    """

    def __init__(self, types=None):
        self.types = types if types else []

        self.count = {}
        self.share = {}

    @property
    def vehicle_types_joinedstr(self):
        """Return the IDs of self.types comma-separated in a single string.
        Example returns:
        ['EN'] -> 'EN'
        ['EN', 'DL'] -> 'EN, DL'
        """
        return ", ".join(vt.ID for vt in self.types)


class SimpleVehicle:
    """Vehicle for the depot simulation.

    Parameters:
    env: [simpy.Environment]
    ID: [str] unique identifier such as vehicle number
    vehicle_type: [VehicleType]

    Attributes:
    trip: [SimpleTrip] current trip of the vehicle. None inside the depot.
    system_entry: [bool] False as long as the vehicle has not been used. True
        after.

    """

    def __init__(self, env, ID, vehicle_type, home_depot=None):
        self.env = env
        self.ID = ID
        self.vehicle_type = vehicle_type
        self.battery = SimpleBattery(
            env,
            vehicle_type.battery_capacity,
            vehicle_type.soc_min,
            vehicle_type.soc_max,
            vehicle_type.soc_init,
            vehicle_type.soh,
        )
        self.mileage = 0
        self.dwd = DepotWorkingData(env, self, home_depot)
        self.trip = None
        self.trip_at_departure = None
        self.finished_trips = []
        self.system_entry = False

        if globalConstants["general"]["LOG_ATTRIBUTES"]:
            self.logger = DataLogger(env, self, "VEHICLE")
            self.logger.loggedData["area_waiting_time"] = {}
            self.logger.loggedData["canceled_precondition"] = {}

        self.battery_logs = []  # Container for BatteryLog objects, temporary
        self.power_logs = {0: 0}  # Container for power logs

    def __repr__(self):
        return "{%s} %s" % (type(self).__name__, self.ID)

    def drive(self):
        """Process one trip.
        Simplifies everything that happens outside of the depot to consuming
        energy and letting time pass if the depot simulations runs on its own.
        Ends by calling checkin when arriving at the depot at the end of a
        trip.
        """
        flexprint(
            "t = %d: %s starting to drive with driving time = %d (= %f h)"
            % (
                self.env.now,
                self.ID,
                self.trip.duration,
                self.trip.duration / 3600,
            ),
            switch="operations",
        )

        self.battery_logs.append(BatteryLog(self.env.now, self, "consume_start"))
        yield self.env.timeout(self.trip.duration)

        if globalConstants["depot"]["consumption_calc_mode"] == "CR_distance_based":
            amount = self.trip.distance * self.vehicle_type.CR
            self.battery.get(amount)
        elif globalConstants["depot"]["consumption_calc_mode"] == "CR_time_based":
            amount = self.trip.duration / 3600 * self.vehicle_type.CR
            self.battery.get(amount)
        elif globalConstants["depot"]["consumption_calc_mode"] == "soc_given":
            if self.trip.charge_on_track:
                self.battery.energy = self.trip.end_soc * self.battery.energy_real
            else:
                used_energy = (
                    self.trip.start_soc - self.trip.end_soc
                ) * self.battery.energy_real
                self.battery.energy -= used_energy
        else:
            raise ValueError(
                "Invalid value %s for 'consumption_calc_mode' in globalConstants."
                % globalConstants["depot"]["consumption_calc_mode"]
            )

        self.battery_logs.append(BatteryLog(self.env.now, self, "consume_end"))

        # Driving time is over, now check in at the depot
        self.mileage += self.trip.distance if self.trip.distance is not None else 0
        self.trip.destination.checkin(self)


class SimpleBattery:
    """Battery for a vehicle.

    Parameters:
    energy_nominal = [float or int: kWh] nominal energy capacity. constant
    soc_min, soc_max: [float] usable energy range. 0...1
    soc_init: [float] initial state of charge within usage energy range
        (energy_real). 0...1
    soh: [float] state of health. 0...1

    Attributes:
    n_charges: number of times the battery has been charged
    active_processes: [list] of BaseDepotProcess or subclass objects that are
        currently modifying energy. These objects must implement and update
        method that modifies energy early.
    last_update: [int] sim time log to prevent recursion when retrieving energy
        and unnecessary updates.

    """

    def __init__(self, env, energy_nominal, soc_min, soc_max, soc_init, soh):
        if energy_nominal <= 0:
            raise ValueError("energy_nominal cannot be <= 0")
        if (
            not soc_input_valid(soc_min)
            or not soc_input_valid(soc_min)
            or not soc_input_valid(soc_min)
            or not soc_input_valid(soc_min)
        ):
            raise ValueError("soc values must be between 0 and 1")

        self.env = env
        self.energy_nominal = energy_nominal
        self.soc_min = soc_min
        self.soc_max = soc_max
        self.soh = soh

        self._energy = soc_init * self.energy_real

        self.n_charges = 0
        self.active_processes = []
        self.updating = False
        self.last_update = 0

    @property
    def energy(self):
        """Return current energy level. Ask active processes for updates first."""
        if self.active_processes and self.last_update != self.env.now:
            self.last_update = self.env.now
            for process in self.active_processes:
                process.update_battery("update")

        return self._energy

    @energy.setter
    def energy(self, value):
        self._energy = value

    @property
    def energy_real(self):
        """Return current real energy capacity."""
        return self.energy_nominal * self.soh

    @property
    def soc(self):
        """Return current state of charge with regard to energy_real."""
        return self.energy / self.energy_real

    @property
    def energy_max(self):
        """Return maximum usable energy."""
        return self.energy_real * self.soc_max

    @property
    def energy_min(self):
        """Return minimum usable energy."""
        return self.energy_real * self.soc_min

    @property
    def energy_remaining(self):
        """Effective amount of energy usable by vehicle at the current state
        of charge."""
        return self.energy - self.energy_min

    def get(self, amount):
        """Subtract *amount* from self.energy.
        If energy would become negative and negativity is not allowed, set it
        to a given positive percentage of self.energy_real. Simple imitation of
        opportunity charging.
        """
        self._energy -= amount

        if self._energy < 0 and not globalConstants["depot"]["allow_negative_soc"]:
            self._energy = (
                globalConstants["depot"]["reset_negative_soc_to"] * self.energy_real
            )

    def put(self, amount):
        """Add *amount* to self.energy."""
        self._energy += amount


def soc_input_valid(soc):
    return 0 <= soc <= 1



"""This module contains miscellaneous utility functions for the eflips-depot API."""

import os
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import timedelta, datetime
from typing import Union, Any, Optional, Tuple, Dict, List

import simpy
from eflips.model import (
    Scenario,
    VehicleType,
    Rotation,
    Event,
    EventType,
    Trip,
    Depot,
)
from sqlalchemy import inspect, create_engine
from sqlalchemy.orm import Session

from eflips.depot import SimpleTrip, Timetable as EflipsTimeTable
from eflips.depot import VehicleType as EflipsVehicleType


@contextmanager
def create_session(
    scenario: Union[Scenario, int, Any], database_url: Optional[str] = None
) -> Tuple[Session, Scenario]:
    """
    Create a valid session from various inputs.

    This method takes a scenario, which can be either a :class:`eflips.model.Scenario` object, an integer specifying
    the ID of a scenario in the database, or any other object that has an attribute `id` that is an integer. It then
    creates a SQLAlchemy session and returns it. If the scenario is a :class:`eflips.model.Scenario` object, the
    session is created and returned. If the scenario is an integer or an object with an `id` attribute, the session
    is created, returned and closed after the context manager is exited.

    :param scenario: Either a :class:`eflips.model.Scenario` object, an integer specifying the ID of a scenario in the
        database, or any other object that has an attribute `id` that is an integer.
    :return: Yield a Tuple of the session and the scenario.
    """
    managed_session = False
    engine = None
    session = None
    try:
        if isinstance(scenario, Scenario):
            session = inspect(scenario).session
        elif isinstance(scenario, int) or hasattr(scenario, "id"):
            if isinstance(scenario, int):
                scenario_id = scenario
            else:
                scenario_id = scenario.id

            if database_url is None:
                if "DATABASE_URL" in os.environ:
                    database_url = os.environ.get("DATABASE_URL")
                else:
                    raise ValueError("No database URL specified.")

            managed_session = True
            engine = create_engine(database_url)
            session = Session(engine)
            scenario = session.query(Scenario).filter(Scenario.id == scenario_id).one()
        else:
            raise ValueError(
                "The scenario parameter must be either a Scenario object, an integer or object with an 'id' attribute."
            )
        yield session, scenario
    finally:
        if managed_session:
            if session is not None:
                session.commit()
                session.close()
            if engine is not None:
                engine.dispose()


def vehicle_type_to_eflips(vt: VehicleType) -> EflipsVehicleType:
    """Convert a VehicleType object to an eflips-depot VehicleType object."""

    # Create the depot VehicleType object
    eflips_vehicle_type = EflipsVehicleType(
        str(vt.id),
        vt.battery_capacity,
        0.0,
        1.0,
        1.0,
        1.0,
        vt.consumption,
    )
    return eflips_vehicle_type


def vehicle_type_to_global_constants_dict(vt: VehicleType) -> Dict[str, float]:
    """
    This converts the VehicleType object into a dictionary, which is the input.

    format of the eflips.globalConstants object.

    :return: A dictionary describing some of the properties of the vehicle type.
    """

    the_dict = {
        "battery_capacity": vt.battery_capacity,
        "soc_min": 0.0,
        "soc_max": 1.0,
        "soc_init": 1.0,
        "soh": 1.0,
    }
    return the_dict


def repeat_vehicle_schedules(
    vehicle_schedules: List["VehicleSchedule"], repetition_period: timedelta
) -> List["VehicleSchedule"]:
    """
    This method repeats the vehicle schedules in the list `vehicle_schedules` by the timedelta `repetition_period`.

    It takes the given vehicle schedules and creates two copies, one `repetition_period` earlier, one
    `repetition_period` later. It then returns the concatenation of the three lists.

    :param vehicle_schedules: A list of :class:`eflips.depot.api.input.VehicleSchedule` objects.
    :param repetition_period: A timedelta object specifying the period of the vehicle schedules.
    :return: a list of :class:`eflips.depot.api.input.VehicleSchedule` objects.
    """
    # Add the repeated schedules to the forward and backward lists
    schedule_list_backward = []
    schedule_list_forward = []

    for vehicle_schedule in vehicle_schedules:
        schedule_list_backward.append(vehicle_schedule.repeat(-repetition_period))
        schedule_list_backward.append(vehicle_schedule.repeat(-2 * repetition_period))

        schedule_list_forward.append(vehicle_schedule.repeat(repetition_period))

    vehicle_schedules = (
        schedule_list_backward + vehicle_schedules + schedule_list_forward
    )

    return vehicle_schedules


def start_and_end_times(vehicle_schedules) -> Tuple[datetime, int]:
    """
    This method is used to find the start time and duration for simulating a given list of vehicle schedules.

    It finds the times of midnight of the day of the first departure and midnight of the day after the last arrival.

    :param vehicle_schedules: A list of :class:`eflips.depot.api.input.VehicleSchedule` objects.
    :return: The datetime of midnight of the day of the first departure and the total duration of the simulation in
        seconds.
    """

    first_departure_time = min(
        [vehicle_schedule.departure for vehicle_schedule in vehicle_schedules]
    )
    midnight_of_first_departure_day = first_departure_time.replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    last_arrival_time = max(
        [vehicle_schedule.arrival for vehicle_schedule in vehicle_schedules]
    )
    midnight_of_last_arrival_day = last_arrival_time.replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    midnight_after_last_arrival_day = midnight_of_last_arrival_day + timedelta(days=1)
    total_duration_seconds = int(
        (
            midnight_after_last_arrival_day - midnight_of_first_departure_day
        ).total_seconds()
    )

    return midnight_of_first_departure_day, total_duration_seconds


def check_depot_validity(depot: Depot) -> None:
    """
    Check if the depot is valid for the eflips-depot simulation.

    Raise an AssertionError if it is not.
    :param depot: a :class:`eflips.model.Depot` object.
    :return: None
    """
    # 1. There must be an area containing no vehicle types
    has_waiting_area = False
    areas = depot.areas
    for area in areas:
        if area.vehicle_type_id is None:
            # TODO might change to len(area.vehicle_types) == 0 after a list of vehicle types is allowed
            has_waiting_area = True

    assert (
        has_waiting_area
    ), "There must be an area containing no vehicle types as the waiting area."

    # 2. There must be only one process with no duration and no electric power, and it must be the last process in the plan
    plan = depot.default_plan
    processes = plan.processes
    last_process = processes[-1]
    assert (
        last_process.duration is None
        and last_process.electric_power is None
        and last_process.dispatchable is True
    ), "The last process must be dispatchable and have no duration and no electric power."

    for process in processes[:-1]:
        assert (
            process.electric_power is not None or process.duration is not None
        ), "All processes except the last one must have electric power."


@dataclass
class VehicleSchedule:
    """
    This class represents a vehicle schedule in eFLIPS-Depot.

    A vehicle schedule presents everything a vehicle does
    between leaving the depot and returning to the depot. In eFLIPS-Depot, we only care about a reduced set of
    information, limited to the interaction with the depot.
    """

    id: str
    """Unique ID of this vehicle schedule.

    This identifier will be returned in the output of eFLIPS-Depot.
    """

    vehicle_type: str
    """The vehicle type of this vehicle schedule.

    This is the ID of a vehicle type in the database.
    """

    departure: datetime
    """
    The departure time of the vehicle from the depot.

    It *must* include the timezone information.
    """

    arrival: datetime
    """
    The arrival time of the vehicle at the depot.

    It *must* include the timezone information.
    """

    departure_soc: float
    """
    The battery state of charge (SoC) of the vehicle at the departure time. It must be in the range [0, 1].

    Note that
    this SoC may not be ctually reached, e.g. if the vehicle is not fully charged when it leaves the depot. The depot
    simulation should always be run multiple times until the `departure_soc` stabilizes.
    """

    arrival_soc: float
    """
    The battery state of charge (SoC) of the vehicles at the arrival time. It must be in the range [-inf, 1].

    This value
    is calculated by a consumption model, e.g. the consumption model of the `ebustoolbox` package. It is a dictionary
    mapping vehicle types to floats. The dictionary must contain an entry for each vehicle type that is part of the
    `vehicle_class` of this vehicle schedule.

    **NOTE**: For the current API version, we only support a single vehicle type per vehicle schedule. This means that
    the dictionary must contain exactly one entry.
    """

    minimal_soc: float
    """
    The minimal battery state of charge (SoC) of the vehicle during the trip.

    It must be in the range [-inf, 1]. This
    value is calculated by a consumption model, e.g. the consumption model of the `ebustoolbox` package.
    """

    opportunity_charging: bool
    """Whether the vehicle is opportunity-charged (meaning charging at terminus stations) during the trip."""

    start_depot_id: str
    """The ID of the depot where the vehicle starts its trip."""

    end_depot_id: str
    """The ID of the depot where the vehicle ends its trip."""

    _is_copy: bool = False
    """
    Whether this vehicle schedule is a copy of another vehicle schedule.

    It should not be set manually, but only by
    calling the :meth:`repeat` method.
    """

    @classmethod
    def from_rotation(
        self,
        rot: Rotation,
        scenario,
        session,
    ):
        """
        This constructor creates a VehicleSchedule object from a Rotation object.

        It is intended to be used by the
        eflips-depot API.

        :param rot: The Rotation object from which the VehicleSchedule is created.
        :param scenario: The Scenario object to which the Rotation belongs.
        :param session: The database session object.
        :param use_builtin_consumption_model: Whether to use the built-in consumption model of eflips-depot. If set
            to `True`, the `VehicleType.consumption` field is used. If set to `False`, consumption is calculated
            from the `Event` table in the database. This (and an external consumption model) is the recommended way.
        """

        rotation_id = str(rot.id)
        departure = rot.trips[0].departure_time
        arrival = rot.trips[-1].arrival_time

        # Find the event for each trip
        events = (
            session.query(Event)
            .filter(Event.event_type == EventType.DRIVING)
            .join(Trip)
            .join(Rotation)
            .filter(Rotation.id == rot.id)
            .order_by(Event.time_start)
            .all()
        )
        trips = (
            session.query(Trip)
            .filter(Trip.rotation_id == rot.id)
            .order_by(Trip.departure_time)
            .all()
        )
        if len(events) != len(trips):
            raise ValueError(
                f"Rotation {rot.id} has {len(trips)} trips but only {len(events)} events."
            )
        if set([event.trip_id for event in events]) != set([trip.id for trip in trips]):
            raise ValueError(f"The events of rotation {rot.id} do not match the trips.")

        departure_soc = events[0].soc_start
        arrival_soc = events[-1].soc_end
        minimal_soc = min([event.soc_end for event in events])

        opportunity_charging = rot.allow_opportunity_charging

        # Find the depot at the start and end of the rotation
        start_depot = trips[0].route.departure_station.depot
        end_depot = trips[-1].route.arrival_station.depot

        if start_depot is None or end_depot is None:
            raise ValueError(f"Rotation {rot.id} has no depot at the start or end.")

        return VehicleSchedule(
            id=rotation_id,
            start_depot_id=str(start_depot.id),
            end_depot_id=str(end_depot.id),
            vehicle_type=str(rot.vehicle_type.id),
            departure=departure,
            arrival=arrival,
            departure_soc=departure_soc,
            arrival_soc=arrival_soc,
            minimal_soc=minimal_soc,
            opportunity_charging=opportunity_charging,
        )

    def _to_simple_trip(
        self, simulation_start_time: datetime, env: simpy.Environment
    ) -> SimpleTrip:
        """
        This converts the vehicle schedule into a :class:`eflips.depot.standalone.SimpleTrip` object, which is the.

        input format of the depot simulation.

        :param simulation_start_time: The time that serves as "zero" for the simulation. It must be before the
            `departure` time of the first of all vehicle schedules, probably midnight of the first day.
        :param env: The simulation environment object. It should be the `env` of the SimulationHost object.

        :return: A :class:`eflips.depot.standalone.SimpleTrip` object.
        """

        vehicle_types = [self.vehicle_type]
        departure = int((self.departure - simulation_start_time).total_seconds())
        arrival = int((self.arrival - simulation_start_time).total_seconds())
        simple_trip = SimpleTrip(
            env=env,
            ID=self.id,
            line_name=None,
            origin=self.start_depot_id,
            destination=self.end_depot_id,
            vehicle_types=vehicle_types,
            std=departure,
            sta=arrival,
            distance=None,
            start_soc=self.departure_soc,
            end_soc=self.arrival_soc,
            charge_on_track=self.opportunity_charging,
            is_copy=self._is_copy,
        )
        return simple_trip

    def repeat(self, interval: timedelta) -> "VehicleSchedule":
        """
        Repeats a given VehicleSchedule.

        Returns a new vehicle schdule offset by a given timedelta that has the
        _copy_of field filled

        :return: A VehicleSchedule object
        """
        sched = VehicleSchedule(
            self.id,
            self.vehicle_type,
            self.departure + interval,
            self.arrival + interval,
            self.departure_soc,
            self.arrival_soc,
            self.minimal_soc,
            self.opportunity_charging,
            start_depot_id=self.start_depot_id,
            end_depot_id=self.end_depot_id,
        )
        sched._is_copy = True
        return sched

    @staticmethod
    def _to_timetable(
        vehicle_schedules: List["VehicleSchedule"],
        env: simpy.Environment,
        start_of_simulation: datetime,
    ) -> EflipsTimeTable:
        """
        This converts a list of VehicleSchedule objects into a :class:`eflips.depot.standalone.Timetable` object, which.

        is the input format of the depot simulation. This Timetable object is part of the "black box" not covered by
        the API documentation.

        :param: vehicle_schedules: A list of :class:`eflips.depot.api.input.VehicleSchedule` objects.
        :param: env: The simulation environment object. It should be the `env` of the SimulationHost object.
        :param: start_of_simulation The datetime that will be used as "zero" for the simulation. It should be before the
            `departure` time of the first of all vehicle schedules, probably midnight of the first day.

        :return: A :class:`eflips.depot.standalone.Timetable` object.
        """

        # Sort the vehicle schedules by departure time
        vehicle_schedules = sorted(vehicle_schedules, key=lambda x: x.departure)

        # Convert the vehicle schedules into SimpleTrip objects
        simple_trips = []
        for vehicle_schedule in vehicle_schedules:
            simple_trip = vehicle_schedule._to_simple_trip(start_of_simulation, env)
            simple_trips.append(simple_trip)

        timetable = EflipsTimeTable(env, simple_trips)

        return timetable



"""Smart charging algorithm."""

import eflips
import eflips.evaluation
from eflips.settings import globalConstants
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import matplotlib.lines as mlines
import math
import datetime


class SmartCharging:
    """
    Class which implements the smart charging algorithm for depot charged vehicles of Lauth, Mundt,
    Göhlich.
    In the eflips_setiings [depot][log_cm_data] has to be True
    """

    def __init__(self, data, start_date, price_data_path, power_limit_grid):
        """

        :param data: data for schedule (DepotEvaluation or pd.Frame
        :param start_date:
        :param price_data_path:
        :param power_limit_grid: int or dict of powerlimits (15 min intervalls) (if list is not long enough, the list starts over again, the last value has to be given for Ex. for a day key 86400 has to exist),
                only the changeges has to be in the dict, index of dict has to be the time in seconds.
         [or an function] not yet implemnted
        """

        if isinstance(
            data, eflips.depot.evaluation.DepotEvaluation
        ):  # for data straight out of eflips
            self.schedule_method(data)
        if isinstance(
            data, pd.core.frame.DataFrame
        ):  # for schedule from externeal programms. Watch that the format is correct.
            self.schedule = data

        self.charging_log = self.construct_empty_dict_charging_log()
        # Outdated, not maintained self.charging_log_even = self.construct_empty_dict_charging_log()
        self.charging_log_imm = self.construct_empty_dict_charging_log()

        self.max_sim_time = self.max_simulation_time()

        try:
            self.start_date = datetime.date(start_date[0], start_date[1], start_date[2])
        except:
            print(
                "The start date has the wrong format. I use the 2019-06-03 as the start date."
            )
            self.start_date = datetime.date(2019, 6, 3)

        power_item = eflips.depot.PowerFrame(
            self.max_sim_time, self.start_date, price_data_path
        )
        self.power = power_item.pdframe()
        self.add_power_limit_grid_to_power(power_limit_grid)

        self.power["used_power_immediately"] = float(0)
        self.power["used_power_smart"] = float(0)
        # Outdated, not maintained self.power["used_power_smart_even"] = float(0)
        if hasattr(self, "precondition_schedule"):
            self.precondition()

        if not (self.immediately_charging()):
            print("Immediately charging, went wrong.")

    def add_power_limit_grid_to_power(self, data):
        if isinstance(data, int):  # backward compatible
            self.power["limit_grid"] = data

        if isinstance(data, dict):
            power_limit_grid = []
            time = 0
            for key, power_limit in data.items():
                while time < int(key):
                    power_limit_grid.append(pre_power_limit)
                    time += 900
                pre_power_limit = power_limit

            while len(self.power.index) > len(power_limit_grid):
                power_limit_grid += power_limit_grid

            # for performance
            self.power["limit_grid"] = power_limit_grid[0 : len(self.power.index)]

    def schedule_method(self, ev):
        """
        Iterates over the process list. And generates the Precon and normal schedule
        :param ev: has to be type DepotEvaluation
        :return: prober schedule and a schdule with all incomplet depot stays
        """

        columns = [
            "charging_power",
            "energy_demand_real",
            "energy_demand",
            "time_arrival",
            "time_depart",
            "time_charging_flex",
            "charging_efficiency",
        ]  # units[kW, kWs,kWs, s, s, s,1] energy_demand_calc is the box integral, which is used for the calculation
        # time depart is always the starting time of the next process
        self.schedule = pd.DataFrame(columns=columns)
        self.schedule_debug = pd.DataFrame(columns=columns)

        precondition_columns = [
            "charging_power",
            "time_start",
            "time_ends",
            "charging_efficiency",
        ]  # units[kW,  s, s, 1]
        self.precondition_schedule = pd.DataFrame(columns=precondition_columns)
        self.vehicles_id = []

        for vehicle in ev.vehicle_generator.items:
            self.vehicles_id.append(vehicle.ID)
            i_charg = 0  # number of charging processes
            i_precon = 0

            pre_process_time = 0
            pre_process_item_charge = None
            pre_precon_process = None

            for process_time, process_list_item in vehicle.logger.loggedData[
                "dwd.active_processes_copy"
            ].items():
                if isinstance(
                    pre_process_item_charge, eflips.depot.processes.ChargeAbstract
                ):
                    row = {}
                    row["time_arrival"] = pre_process_item_charge.starts[0]
                    row[
                        "time_depart"
                    ] = process_time  # Depart time of the process before is the estart time of the next process

                    charging_duration = self.calc_charging_time(
                        pre_process_item_charge.starts, pre_process_item_charge.ends
                    )
                    row["time_charging_flex"] = (
                        row["time_depart"] - row["time_arrival"] - charging_duration
                    )

                    row["energy_demand_real"] = (
                        pre_process_item_charge.energy * 3600
                    )  # kWh ->kWs
                    row["charging_efficiency"] = pre_process_item_charge.efficiency

                    row["charging_power"] = vehicle.power_logs[row["time_arrival"]]
                    row["energy_demand"] = charging_duration * row["charging_power"]

                    if (
                        row["energy_demand_real"] > row["energy_demand"]
                    ):  # Makes sure the that the numerical error gets deleted
                        row["energy_demand_real"] = row["energy_demand"]

                    all_keys_exist = True
                    for key_to_check in columns:  # checks if dict is complet
                        if key_to_check in row:
                            all_keys_exist = True
                        else:
                            all_keys_exist = False
                            break

                    if all_keys_exist:  # safes the row in the dataframe
                        row = pd.Series(row)
                        self.schedule.loc[vehicle.ID + "_" + str(i_charg)] = row
                        if row["energy_demand"] / row["energy_demand_real"] > 2:
                            print(
                                "Bus "
                                + vehicle.ID
                                + " has deviation of over 50 percent between energy_demand_real and energy_demand. You might wanna check this."
                            )
                    else:
                        row = pd.Series(row)
                        self.schedule_debug.loc[vehicle.ID + "_" + str(i_charg)] = row

                    i_charg += 1
                    pre_process_item_charge = None  # Reset the item

                for process_item in process_list_item:
                    if isinstance(process_item, eflips.depot.processes.ChargeAbstract):
                        # Filter and Store the correct processes
                        pre_process_item_charge = process_item

                    if (
                        isinstance(process_item, eflips.depot.processes.Precondition)
                        and process_item is not pre_precon_process
                    ):
                        # Filter and Store the correct processes
                        row = {
                            "charging_power": process_item.power,
                            "time_start": process_item.starts[0],
                        }

                        if process_item.ends:  # Checks if Precon process is complete
                            row["time_ends"] = process_item.ends[0]
                            row["charging_efficiency"] = process_item.efficiency

                            row = pd.Series(row)
                            self.precondition_schedule.loc[
                                vehicle.ID + "_" + str(i_precon)
                            ] = row

                            i_precon += 1

                            pre_precon_process = process_item  # Because if a precon process is canceled, it is logged twice. It has to be checked if the precon process was already used.

        print("Schedule building done.")

    def calc_charging_time(self, start, end):
        """

        :param start: a list of all start times
        :param end: a list of all end times
        :return: total duration of charging
        """
        duration = 0
        for i in range(len(start)):
            if i <= len(end):
                duration += end[i] - start[i]

        return duration

    def set_power(self, power):
        """
        :param power: a pd_frame
        """
        self.power = power

    def max_simulation_time(self):
        """

        :return: The maximum time.
        """
        return self.schedule.max().time_depart

    def construct_pd_frame_charging_history_energy(self):
        """

        :return: empty pd Frame with time as index and bus stops in depot as columns
        """
        columns = list(self.schedule.index)
        index = list(self.power.index)
        charging_history = pd.DataFrame(index=index, columns=columns)
        return charging_history

    def construct_empty_dict_charging_log(self):
        """

        :return: a dict, in which al the charging intervalles of the vehicles will be stored
        """
        columns = ["start", "end", "power", "price"]
        dict_of_pd = {}
        for ID in list(self.schedule.index):
            # words = ID.split("_")
            # ID_bus = words[0] + "_" + words[1]
            dict_of_pd[ID] = pd.DataFrame(columns=columns)

        return dict_of_pd

    def smart_charging_algorithm(self):
        """
        needs self.schedule, self.power_price (start_time, end_time, price), power_limi_grid. Implements the algorithm of Lauth, Mundt,
        Göhlich.
        :return: False if smart charging is not possible, a pd Frame with the time intervalls and the charging power, if smart charging was possible
        """

        schedule = self.schedule.sort_values(by=["time_charging_flex"])

        self.power = self.power.sort_values(by=["price"])
        for ID_bus, row_bus in schedule.iterrows():
            # power delta is needed for later adjustments, to match the real energy need
            energy_delta = row_bus.energy_demand - row_bus.energy_demand_real

            for ID_power, row_power in self.power.iterrows():
                if row_bus.energy_demand <= 0:
                    break
                power_rest = (
                    row_power.limit_grid - row_power.used_power_smart
                )  # determines the left power
                interval = self.intersection(
                    [row_power.start_time, row_power.end_time],
                    [row_bus.time_arrival, row_bus.time_depart],
                )
                if isinstance(interval, list) and power_rest > 0:
                    self.energy_calculation_smart(
                        interval[0],
                        interval[1],
                        row_bus,
                        row_power,
                        ID_power,
                        power_rest,
                        ID_bus,
                    )

            if row_bus.energy_demand >= 50:  # needed against numerical error
                print(row_bus.energy_demand, ID_bus, " Smart")
                return False

            # Make adjustments so the charging energy fit the real energy
            if energy_delta > 0:
                self.compensate_to_energy_real(
                    energy_delta, ID_bus, row_bus, self.charging_log, "used_power_smart"
                )

        return True

    def energy_calculation_smart(
        self, start_time, end_time, row_bus, row_power, ID_power, power_rest, ID_bus
    ):
        """
        For outsource some of the stuff of smart_charging_algorithm
        :param start_time:
        :param end_time:
        :param row_bus:
        :param ID_power:
        :param power_rest:
        :return:
        """
        factor_power_used = (end_time - start_time) / (
            row_power.end_time - row_power.start_time
        )  # a factor which scales the power, if a bus intervall is not complet in a power intervall

        if power_rest >= row_bus.charging_power / row_bus.charging_efficiency:
            if (
                row_bus.energy_demand
                >= (end_time - start_time) * row_bus.charging_power
            ):
                energy = (end_time - start_time) * row_bus.charging_power
                row_bus.energy_demand -= energy
                # self.energy_log.at[row_power.start_time,ID_bus] = energy

                power = (
                    row_bus.charging_power / row_bus.charging_efficiency
                ) * factor_power_used
                self.power.at[ID_power, "used_power_smart"] += power
                row = [
                    start_time,
                    end_time,
                    power * row_bus.charging_efficiency,
                    row_power.price,
                ]
                self.charging_log[ID_bus].loc[int(start_time)] = row
            else:
                power = (
                    row_bus.energy_demand
                    / ((end_time - start_time) * row_bus.charging_efficiency)
                    * factor_power_used
                )
                self.power.at[ID_power, "used_power_smart"] += power
                row = [
                    start_time,
                    end_time,
                    power * row_bus.charging_efficiency,
                    row_power.price,
                ]
                self.charging_log[ID_bus].loc[int(start_time)] = row

                # self.energy_log.at[row_power.start_time, ID_bus] = row_bus.energy_demand
                row_bus.energy_demand = 0
        else:
            if (
                row_bus.energy_demand
                >= (end_time - start_time) * power_rest * row_bus.charging_efficiency
            ):
                energy = (
                    (end_time - start_time) * power_rest * row_bus.charging_efficiency
                )
                row_bus.energy_demand -= energy
                # self.energy_log.at[row_power.start_time, ID_bus] = energy

                power = power_rest * factor_power_used
                self.power.at[ID_power, "used_power_smart"] += power
                row = [
                    start_time,
                    end_time,
                    power * row_bus.charging_efficiency,
                    row_power.price,
                ]
                self.charging_log[ID_bus].loc[int(start_time)] = row
            else:
                power = (
                    row_bus.energy_demand
                    / ((end_time - start_time) * row_bus.charging_efficiency)
                ) * factor_power_used
                self.power.at[ID_power, "used_power_smart"] += power
                row = [
                    start_time,
                    end_time,
                    power * row_bus.charging_efficiency,
                    row_power.price,
                ]
                self.charging_log[ID_bus].loc[int(start_time)] = row

                # self.energy_log.at[row_power.start_time, ID_bus] = row_bus.energy_demand
                row_bus.energy_demand = 0

        return row_bus.energy_demand

    def smart_charging_algorithm_even(self):  # Outdated, not maintained
        """Same power curve, just buses will always be charged."""

        for ID_bus, row_bus in self.schedule.iterrows():
            constant_power = row_bus.energy_demand / (
                row_bus.time_depart - row_bus.time_arrival
            )

            for ID_power, row_power in self.power.iterrows():
                interval = self.intersection(
                    [row_power.start_time, row_power.end_time],
                    [row_bus.time_arrival, row_bus.time_depart],
                )
                if isinstance(interval, list):
                    self.energy_calculation_smart_even(
                        interval[0],
                        interval[1],
                        row_bus,
                        row_power,
                        ID_power,
                        ID_bus,
                        constant_power,
                    )

        self.power = self.power.sort_values(by=["price"])

        for ID_power, row_power in self.power.iterrows():
            if row_power.used_power_smart_even > 0:
                scaling_quotients = (
                    row_power.limit_grid / row_power.used_power_smart_even
                )
                if scaling_quotients > 1:
                    self.smart_even_upscalinig(scaling_quotients, ID_power)
                elif scaling_quotients < 1:
                    self.smart_even_downscalinig(scaling_quotients, ID_power)

        if any(
            self.power.used_power_smart_even > self.power.limit_grid * 1.01
        ):  # *1.01 because of the inaccuracy of the numeric calculations, in the calculations before
            print("Smart even failed")
            return False
        else:
            return True

    def energy_calculation_smart_even(
        self, start_time, end_time, row_bus, row_power, ID_power, ID_bus, constant_power
    ):  # Outdated, not maintained
        factor_power_used = (end_time - start_time) / (
            row_power.end_time - row_power.start_time
        )  # a factor which scales the power, if a bus intervall is not complet in a power intervall

        power = (constant_power * factor_power_used) / row_bus.charging_efficiency
        self.power.at[ID_power, "used_power_smart_even"] += power
        row = [row_power.start_time, row_power.end_time, power, row_power.price]
        self.charging_log_even[ID_bus].loc[int(row_power.start_time)] = row

    def smart_even_upscalinig(
        self, scaling_quotients, ID_power
    ):  # Outdated, not maintained
        """Scales the power for a specific intervall by scaling quotinet up. Also starts the proces to substract the power from an other power intervall."""

        for ID_bus, value in self.charging_log_even.items():
            for start_time, row in value.iterrows():
                if ID_power == start_time:
                    # Some checks
                    if (
                        row.power * scaling_quotients
                        > self.schedule.at[ID_bus, "charging_power"]
                    ):
                        power_delta = (
                            self.schedule.at[ID_bus, "charging_power"] - row.power
                        )
                        self.power.at[ID_power, "used_power_smart_even"] += power_delta
                        self.substract_power_most_expensive(
                            power_delta, ID_bus, ID_power
                        )
                    else:
                        power_delta = (row.power * scaling_quotients) - row.power
                        self.power.at[ID_power, "used_power_smart_even"] += power_delta
                        self.substract_power_most_expensive(
                            power_delta, ID_bus, ID_power
                        )

    def substract_power_most_expensive(
        self, power, ID_bus, ID_power
    ):  # Outdated, not maintained
        """Substracts the power from an other power intervall."""
        self.charging_log_even[ID_bus] = self.charging_log_even[ID_bus].sort_values(
            by=["price"], ascending=False
        )
        for start_time, row in self.charging_log_even[ID_bus].iterrows():
            if start_time != ID_power:
                if row.power >= power:
                    self.charging_log_even[ID_bus].at[start_time, "power"] -= power
                    self.power.at[start_time, "used_power_smart_even"] -= power
                    break
                else:
                    power -= self.charging_log_even[ID_bus].at[start_time, "power"]
                    self.power.at[
                        start_time, "used_power_smart_even"
                    ] -= self.charging_log_even[ID_bus].at[start_time, "power"]
                    self.charging_log_even[ID_bus].at[start_time, "power"] = 0

    def smart_even_downscalinig(
        self, scaling_quotients, ID_power
    ):  # Outdated, not maintained
        """Scales the power for a specific intervall by scaling quotinet down. Also starts the proces to add the power from an other power intervall."""
        for ID_bus, value in self.charging_log_even.items():
            for start_time, row in value.iterrows():
                if ID_power == start_time:
                    power_delta = row.power - (row.power * scaling_quotients)
                    self.power.at[ID_power, "used_power_smart_even"] -= power_delta
                    self.add_power_less_expensive(power_delta, ID_bus, ID_power)

    def add_power_less_expensive(
        self, power, ID_bus, ID_power
    ):  # Outdated, not maintained
        # Adds the power to an other intervall.
        self.charging_log_even[ID_bus] = self.charging_log_even[ID_bus].sort_values(
            by=["price"]
        )
        for start_time, row in self.charging_log_even[ID_bus].iterrows():
            if start_time != ID_power:
                potential_power = min(
                    [
                        self.schedule.at[ID_bus, "charging_power"],
                        self.power.at[start_time, "limit_grid"]
                        - self.power.at[start_time, "used_power_smart_even"],
                    ]
                )  # max power of charging infrastructur of a singele bus, left power until max of the grid
                if potential_power >= power:
                    self.charging_log_even[ID_bus].at[start_time, "power"] += power
                    self.power.at[start_time, "used_power_smart_even"] += power
                    break
                else:
                    self.charging_log_even[ID_bus].at[
                        start_time, "power"
                    ] += potential_power
                    self.power.at[
                        start_time, "used_power_smart_even"
                    ] += potential_power
                    power -= potential_power

    def immediately_charging(self):
        """
        Charges the buses immediatley after the parked. No charging eqauation steps.
        :return: pd Frame
        """

        schedule = self.schedule.copy()

        self.power = self.power.sort_values(by=["start_time"])
        for ID_bus, row_bus in schedule.iterrows():
            # power delta is needed for later adjustments, to match the real energy need
            energy_delta = row_bus.energy_demand - row_bus.energy_demand_real

            for ID_power, row_power in self.power.iterrows():
                interval = self.intersection(
                    [row_power.start_time, row_power.end_time],
                    [row_bus.time_arrival, row_bus.time_depart],
                )
                if isinstance(interval, list):
                    self.energy_calculation_immediately(
                        interval[0], interval[1], row_bus, row_power, ID_power, ID_bus
                    )

            if row_bus.energy_demand >= 50:  # needed against numerical error
                print(row_bus.energy_demand, ID_bus, "Imm")
                return False

            # Make adjustments so the charging energy fit the real energy
            if energy_delta > 0:
                self.compensate_to_energy_real(
                    energy_delta,
                    ID_bus,
                    row_bus,
                    self.charging_log_imm,
                    "used_power_immediately",
                )

        print("Immediately charging done.")
        return True

    def energy_calculation_immediately(
        self, start_time, end_time, row_bus, row_power, ID_power, ID_bus
    ):
        """
        Helper
        :param start_time:
        :param end_time:
        :param row_bus:
        :param row_power:
        :param ID_power:
        :return:
        """

        factor_power_used = (end_time - start_time) / (
            row_power.end_time - row_power.start_time
        )  # a factor which scales the power, if a bus intervall is not complet in a power intervall

        if row_bus.energy_demand >= (end_time - start_time) * row_bus.charging_power:
            row_bus.energy_demand -= (end_time - start_time) * row_bus.charging_power
            self.power.at[ID_power, "used_power_immediately"] += (
                row_bus.charging_power / row_bus.charging_efficiency
            ) * factor_power_used
            row = [start_time, end_time, row_bus.charging_power, row_power.price]
            self.charging_log_imm[ID_bus].loc[int(start_time)] = row
        else:
            power_at_grid = (
                row_bus.energy_demand
                / ((end_time - start_time) * row_bus.charging_efficiency)
            ) * factor_power_used
            self.power.at[ID_power, "used_power_immediately"] += power_at_grid
            row_bus.energy_demand = 0
            row = [
                start_time,
                end_time,
                power_at_grid * row_bus.charging_efficiency,
                row_power.price,
            ]
            self.charging_log_imm[ID_bus].loc[int(start_time)] = row

        return row_bus.energy_demand

    def intersection(self, interval_a, interval_b):
        """

        :param interval_a: list of two numbers
        :param interval_b: list of two numbers
        :return: the intersection of the two intervalls or False if no intersection is there
        """
        lower = max(interval_a[0], interval_b[0])
        upper = min(interval_a[1], interval_b[1])
        if lower >= upper:
            return False
        else:
            return [lower, upper]

    def in_period(self, interval, time):
        """Checks if time is in interval."""
        if interval[0] < interval[1]:
            return interval[0] <= time <= interval[1]
        else:
            return interval[0] <= time <= 86400 or time <= interval[1]

        return False

    def compensate_to_energy_real(
        self, energy_delta, ID_bus, row_bus, charging_log, type_of_power
    ):
        """
        Makes the correct adjustments, so charg_eqaution_steps can be implemented.
        :param energy_delta: differnce between energy and energy_real
        :param ID_bus:
        :param row_bus:
        :param charging_log: the charging log the function should use
        type_of_power: "used_power_smart" or "used_power_immediately"
        :return:
        """

        charging_log[ID_bus] = charging_log[ID_bus].sort_values(
            by="start", ascending=False
        )
        for start_time, charg_log in charging_log[ID_bus].iterrows():
            energy_slot = (charg_log.end - charg_log.start) * charg_log.power

            if energy_delta >= energy_slot:
                energy_delta -= energy_slot

                # find the correct time slot in power and substract the power
                power_time = int(start_time / 900) * 900

                self.power.at[power_time, type_of_power] -= (
                    charg_log.power / row_bus.charging_efficiency
                )

                charging_log[ID_bus].at[start_time, "power"] = 0

            else:
                energy_slot = energy_slot - energy_delta
                energy_delta = 0

                # find the correct time slot in power and substract the power
                power_time = int(start_time / 900) * 900
                power_at_bus = energy_slot / (charg_log.end - charg_log.start)

                self.power.at[power_time, type_of_power] -= (
                    charg_log.power - power_at_bus
                ) / row_bus.charging_efficiency
                charging_log[ID_bus].at[start_time, "power"] = power_at_bus

            if energy_delta <= 0:
                break

    def precondition(self):
        """Writes only the power in the power DF not in the charging logs. Writes the power always in all three charging type columns"""

        self.power = self.power.sort_values(by=["start_time"])
        self.power["precondition"] = float(0)
        for ID_bus, row_bus in self.precondition_schedule.iterrows():
            for ID_power, row_power in self.power.iterrows():
                interval = self.intersection(
                    [row_power.start_time, row_power.end_time],
                    [row_bus.time_start, row_bus.time_ends],
                )
                if isinstance(interval, list):
                    factor_power_used = (interval[1] - interval[0]) / (
                        row_power.end_time - row_power.start_time
                    )
                    power = (
                        row_bus.charging_power / row_bus.charging_efficiency
                    ) * factor_power_used
                    self.power.at[ID_power, "used_power_immediately"] += power
                    self.power.at[ID_power, "used_power_smart"] += power
                    # self.power.at[ID_power,"used_power_smart_even"] += power
                    self.power.at[ID_power, "precondition"] += power

    def plot_results(self, language="eng"):
        """
        Plots the results
        :return:
        """
        power = self.power.sort_values(by=["start_time"])

        x = list(power.start_time)

        used_power_immediately = list(power.used_power_immediately)
        if hasattr(self, "precondition_schedule"):
            precondition = list(power.precondition)
        power_limit_grid = list(power.limit_grid)
        price = list(power.price)
        y_max = max(used_power_immediately) + 2000

        fig, ax1 = plt.subplots(figsize=(7, 4.5))
        eflips.depot.evaluation.setting_language(language)

        if language == "eng":
            ax1.set_xlabel("Time")
        else:
            ax1.set_xlabel("Zeit")

        ax1.set_title(
            "Charging power (all buses) and power price, Date: " + str(self.start_date)
        )

        # Convert x axis seconds to dates
        eflips.depot.evaluation.to_dateaxis(ax1)

        # plot used_power_smart
        if "used_power_smart" in power:
            used_power_smart = list(power.used_power_smart)
            ax1.set_ylim(0, y_max)
            if language == "eng":
                ax1.set_ylabel("Power [kW]")
            else:
                ax1.set_ylabel("Leistung [kW]")
            ax1.yaxis.grid(True)
            plt.fill_between(x, used_power_smart, 0, color="grey", alpha=0.7)

        # plot used_power_smart_even
        if "used_power_smart_even" in power:
            used_power_smart_even = list(power.used_power_smart_even)
            plt.plot(x, used_power_smart_even)

        # plot used_power_immediately
        plt.fill_between(x, used_power_immediately, 0, color="grey", alpha=0.3)

        # plot precondition
        if language == "eng":
            (precondition,) = ax1.plot(
                x,
                precondition,
                color="k",
                linestyle="--",
                linewidth=1,
                label="precondition",
            )
        else:
            (precondition,) = ax1.plot(
                x,
                precondition,
                color="k",
                linestyle="--",
                linewidth=1,
                label="Vorkonditionierung",
            )

        # plot power limit grid
        if language == "eng":
            (power_limit_grid,) = ax1.plot(
                x, power_limit_grid, color="k", linewidth=1, label="power limit grid"
            )
        else:
            (power_limit_grid,) = ax1.plot(
                x,
                power_limit_grid,
                color="k",
                linewidth=1,
                label="Limit Gesamtladeleistung",
            )

        # if language == 'eng':
        #     black_line = mlines.Line2D([], [], color='k', label='power limit grid')
        # else:
        #     black_line = mlines.Line2D([], [], color='k', label='power limit grid')

        # plot price
        ax2 = ax1.twinx()
        if language == "eng":
            (spot_price,) = ax2.plot(
                x, price, color="r", linewidth=1, label="spot price"
            )
        else:
            (spot_price,) = ax2.plot(
                x, price, color="r", linewidth=1, label="Preis Spotmarkt"
            )
        if language == "eng":
            ax2.set_ylabel("Spot price [€/kWh]", color="r")
        else:
            ax2.set_ylabel("Preis Spotmarkt [€/kWh]", color="r")
        ax2.tick_params("y", colors="r")
        ax2.set_ylim(0, max(price) + 0.02)

        if language == "eng":
            lightgrey_patch = mpatches.Patch(
                color="grey", label="non-controlled charging", alpha=0.3
            )
        else:
            lightgrey_patch = mpatches.Patch(
                color="grey", label="Lademanagement: Nicht gesteuert", alpha=0.3
            )
        if language == "eng":
            grey_patch = mpatches.Patch(color="grey", label="smart charging", alpha=0.7)
        else:
            grey_patch = mpatches.Patch(
                color="grey", label="Lademanagement: Smart", alpha=0.7
            )

        plt.legend(
            handles=[
                lightgrey_patch,
                grey_patch,
                power_limit_grid,
                precondition,
                spot_price,
            ],
            mode="expand",
            ncol=2,
        )

        plt.show()
        fig.savefig(
            globalConstants["depot"]["path_results"]
            + "charging_profile_non-controlled_vs_smart.pdf"
        )

    def plot_charging_intervalls_smart(self):
        self.plot_charging_intervalls(self.charging_log)

    def plot_charging_intervalls_smart_even(self):
        self.plot_charging_intervalls(self.charging_log_even)

    def plot_charging_intervalls(self, dict_charging_logs):
        """
        Plots a figure, where you can see the stress of the battries.
        :return:
        """

        # combines the diffrent stays of a vehicle to one "vehicle"
        dict_of_pd_plot = {}
        for key, value in dict_charging_logs.items():
            ID_bus = ""
            words = key.split("_")
            for i in range(len(words) - 1):
                if i == 0:
                    ID_bus += words[i]
                else:
                    ID_bus += "_" + words[i]
            try:
                pre_pd = dict_of_pd_plot[ID_bus]
            except:
                pre_pd = None
            dict_of_pd_plot[ID_bus] = pd.concat([pre_pd, value])

        fig2, ax10 = plt.subplots()
        ax10.set_xlabel("Time (hh:mm)\n Date (yyyy-mm-dd)")
        ax10.set_title("Charging time for each bus")

        # x Tick labels
        days = math.ceil(self.max_sim_time / 86400)
        current_day = self.start_date
        x_labels = []
        x_position = []
        for day in range(days):
            x_labels.append("00:00\n" + str(current_day))
            current_day += datetime.timedelta(days=1)
            x_position.append(day * 86400)
            x_labels.append("12:00")
            x_position.append(day * 86400 + 43200)
        ax10.set_xticks(x_position)
        ax10.set_xticklabels(x_labels)

        y_labels = []
        y_position_bar = 0
        for key, value in dict_of_pd_plot.items():
            y_position_bar += 5
            y_labels.append(key)
            broken_bar = []
            for index, row in value.iterrows():
                broken_bar.append([row.start, row.end - row.start])
                ax10.broken_barh(broken_bar, (y_position_bar, row.power * 0.027))

        ax10.set_yticks(list(range(5, y_position_bar, 5)))
        ax10.set_yticklabels(y_labels, fontdict={"fontsize": 4})

        plt.show

    def results(self):
        """
        Calculates some results and prints them.
        :return:
        """
        columns = ["Imm. Charging", "Smart Charging", "Comparison [%]"]
        results = pd.DataFrame(columns=columns)

        power_sorted_interval = self.power.sort_index()
        for index, row in power_sorted_interval.iterrows():
            if index < 86400 or index >= 172800:
                power_sorted_interval.drop([index], inplace=True)

        if "used_power_immediately" in self.power.columns:
            max_power_ic = round(max(self.power.used_power_immediately))
            results.at["max. Power (kW)", "Imm. Charging"] = max_power_ic

            energy_ic = round(
                (
                    self.power.used_power_immediately
                    * (self.power.end_time - self.power.start_time)
                    / 3600
                ).sum()
            )
            results.at["Energy (kWh)", "Imm. Charging"] = energy_ic

            energy_ic_interval = round(
                (
                    power_sorted_interval.used_power_immediately
                    * (
                        power_sorted_interval.end_time
                        - power_sorted_interval.start_time
                    )
                    / 3600
                ).sum()
            )
            results.at["Energy Interval (kWh)", "Imm. Charging"] = energy_ic_interval

            # energy_ic_cost_spot = round(((self.power.used_power_immediately * (self.power.end_time - self.power.start_time) / 3600) * self.power.price).sum(), 2)
            # results.at["Energy Costs Spot (€)", "Imm. Charging"] = energy_ic_cost_spot

            energy_ic_cost_fix = round(
                (
                    (
                        self.power.used_power_immediately
                        * (self.power.end_time - self.power.start_time)
                        / 3600
                    )
                    * 0.0433
                ).sum(),
                2,
            )
            results.at["Energy Costs (€)", "Imm. Charging"] = energy_ic_cost_fix

            energy_ic_cost_fix_interval = round(
                (
                    (
                        power_sorted_interval.used_power_immediately
                        * (
                            power_sorted_interval.end_time
                            - power_sorted_interval.start_time
                        )
                        / 3600
                    )
                    * 0.0433
                ).sum(),
                2,
            )
            results.at[
                "Energy Costs Interval (€)", "Imm. Charging"
            ] = energy_ic_cost_fix_interval

            grid_ic_cost = round(
                max_power_ic * 66.82 / 365 * self.max_sim_time / (24 * 60 * 60)
                + energy_ic * 0.0098,
                2,
            )
            results.at["Grid Costs (€)", "Imm. Charging"] = grid_ic_cost

            grid_ic_cost_interval = round(
                max_power_ic * 66.82 / 365 + energy_ic_interval * 0.0098, 2
            )
            results.at[
                "Grid Costs Interval (€)", "Imm. Charging"
            ] = grid_ic_cost_interval

            usage_per_year_ic = round(
                energy_ic / max_power_ic / self.max_sim_time * 365 * 24 * 60 * 60
            )
            results.at["Usage Hours per Year (h)", "Imm. Charging"] = usage_per_year_ic

            usage_per_year_ic_interval = round(energy_ic_interval / max_power_ic * 365)
            results.at[
                "Usage Hours per Year Interval (h)", "Imm. Charging"
            ] = usage_per_year_ic_interval

            usage_per_day_ic_interval = round(energy_ic_interval / max_power_ic, 2)
            results.at[
                "Usage Hours per Day Interval (h)", "Imm. Charging"
            ] = usage_per_day_ic_interval

        if "used_power_smart" in self.power.columns:
            max_power_sc = round(max(self.power.used_power_smart))
            results.at["max. Power (kW)", "Smart Charging"] = max_power_sc

            energy_sc = round(
                (
                    self.power.used_power_smart
                    * (self.power.end_time - self.power.start_time)
                    / 3600
                ).sum()
            )
            results.at["Energy (kWh)", "Smart Charging"] = energy_sc

            energy_sc_interval = round(
                (
                    power_sorted_interval.used_power_smart
                    * (
                        power_sorted_interval.end_time
                        - power_sorted_interval.start_time
                    )
                    / 3600
                ).sum()
            )
            results.at["Energy Interval (kWh)", "Smart Charging"] = energy_sc_interval

            energy_sc_cost_spot = round(
                (
                    (
                        self.power.used_power_smart
                        * (self.power.end_time - self.power.start_time)
                        / 3600
                    )
                    * self.power.price
                ).sum(),
                2,
            )
            results.at["Energy Costs (€)", "Smart Charging"] = energy_sc_cost_spot

            energy_sc_cost_spot_interval = round(
                (
                    (
                        power_sorted_interval.used_power_smart
                        * (
                            power_sorted_interval.end_time
                            - power_sorted_interval.start_time
                        )
                        / 3600
                    )
                    * power_sorted_interval.price
                ).sum(),
                2,
            )
            results.at[
                "Energy Costs Interval (€)", "Smart Charging"
            ] = energy_sc_cost_spot_interval

            # energy_sc_cost_fix = round(((self.power.used_power_smart * (self.power.end_time - self.power.start_time) / 3600) * 4.33).sum(), 2)
            # results.at["Energy Costs Fix Price (€)", "Smart Charging"] = energy_sc_cost_fix

            grid_sc_cost = round(
                max_power_sc * 66.82 / 365 * self.max_sim_time / (24 * 60 * 60)
                + energy_sc * 0.0098,
                2,
            )
            results.at["Grid Costs (€)", "Smart Charging"] = grid_sc_cost

            grid_sc_cost_interval = round(
                max_power_sc * 66.82 / 365 + energy_sc_interval * 0.0098, 2
            )
            results.at[
                "Grid Costs Interval (€)", "Smart Charging"
            ] = grid_sc_cost_interval

            usage_per_year_sc = round(
                energy_sc / max_power_sc / self.max_sim_time * 365 * 24 * 60 * 60
            )
            results.at["Usage Hours per Year (h)", "Smart Charging"] = usage_per_year_sc

            usage_per_year_sc_interval = round(energy_sc_interval / max_power_sc * 365)
            results.at[
                "Usage Hours per Year Interval (h)", "Smart Charging"
            ] = usage_per_year_sc_interval

            usage_per_day_sc_interval = round(energy_sc_interval / max_power_sc, 2)
            results.at[
                "Usage Hours per Day Interval (h)", "Smart Charging"
            ] = usage_per_day_sc_interval

        if "used_power_smart" and "used_power_immediately" in self.power.columns:
            results.at["max. Power (kW)", "Comparison [%]"] = round(
                (max_power_sc / max_power_ic - 1) * 100, 2
            )
            results.at["Energy (kWh)", "Comparison [%]"] = round(
                (energy_sc / energy_ic - 1) * 100, 2
            )
            results.at["Energy Interval (kWh)", "Comparison [%]"] = round(
                (energy_sc_interval / energy_ic_interval - 1) * 100, 2
            )
            results.at["Energy Costs (€)", "Comparison [%]"] = round(
                (energy_sc_cost_spot / energy_ic_cost_fix - 1) * 100, 2
            )
            results.at["Energy Costs Interval (€)", "Comparison [%]"] = round(
                (energy_sc_cost_spot_interval / energy_ic_cost_fix_interval - 1) * 100,
                2,
            )
            # results.at["Energy Costs Fix Price (€)", "Comparison [%]"] = round((energy_sc_cost_fix / energy_ic_cost_fix - 1) * 100, 2)
            results.at["Grid Costs (€)", "Comparison [%]"] = round(
                (grid_sc_cost / grid_ic_cost - 1) * 100, 2
            )
            results.at["Grid Costs Interval (€)", "Comparison [%]"] = round(
                (grid_sc_cost_interval / grid_ic_cost_interval - 1) * 100, 2
            )
            results.at["Usage Hours per Year (h)", "Comparison [%]"] = round(
                (usage_per_year_sc / usage_per_year_ic - 1) * 100, 2
            )
            results.at["Usage Hours per Year Interval (h)", "Comparison [%]"] = round(
                (usage_per_year_sc_interval / usage_per_year_ic_interval - 1) * 100, 2
            )
            results.at["Usage Hours per Day Interval (h)", "Comparison [%]"] = round(
                (usage_per_day_sc_interval / usage_per_day_ic_interval - 1) * 100, 2
            )

        # Outdated, not maintained
        """if "used_power_smart_even" in self.power.columns:
            results.at["Smart Charging Even", "Energy (kWh)"] = round((self.power.used_power_smart_even * ( self.power.end_time - self.power.start_time) / 3600).sum(),2)
            results.at["Smart Charging Even", "Cost of Energy (€)"] = round(((
                        self.power.used_power_smart_even * (self.power.end_time - self.power.start_time) / 3600) * self.power.price).sum(),2)
            results.at["Smart Charging Even", "Period of use (min)"] = self.period_of_use("used_power_smart_even")"""

        # self.results = results

        return results

    def period_of_use(self, type):
        """
        Calculates the time [minutes] which the grid infrastructure is used.
        :param type:
        :return:
        """
        count = self.power.loc[self.power[type] > 0].count().start_time
        return count * 15

    def export_for_vehicle_periods_smart(self):
        return self.export_for_vehicle_periods(self.charging_log)

    def export_for_vehicle_periods_smart_even(self):  # Outdated, not maintained
        return self.export_for_vehicle_periods(self.charging_log_even)

    def export_for_vehicle_periods(self, dict_charging_logs):
        """Brings the data to an format, evalutatio.DepotEvaluation.vehicle_periods can work with."""

        dict_of_pd_plot = {}
        for key, value in dict_charging_logs.items():
            ID_bus = ""
            words = key.split("_")
            for i in range(len(words) - 1):
                if i == 0:
                    ID_bus += words[i]
                else:
                    ID_bus += "_" + words[i]

            try:
                pre_pd = dict_of_pd_plot[ID_bus]
            except:
                pre_pd = None
            dict_of_pd_plot[ID_bus] = pd.concat([pre_pd, value])

        vehicle_periods = {}
        i = 0
        for vehicle_id in self.vehicles_id:
            i += 1
            vehicle_periods[vehicle_id] = {"xranges": [], "yranges": (i, 0.5)}
            if vehicle_id in dict_of_pd_plot:
                for index, row in dict_of_pd_plot[vehicle_id].iterrows():
                    vehicle_periods[vehicle_id]["xranges"].append(
                        (row.start, row.end - row.start)
                    )

        return vehicle_periods

    def validation(self):
        """Validates some values, and prints weired occurrences, which might be checked by the user"""

        results = self.results()
        valid = True
        total_energy_real = self.schedule["energy_demand_real"].sum() / 3600

        # Check precondition durations
        if (
            self.precondition_schedule["time_ends"]
            - self.precondition_schedule["time_start"]
        ).min() < 0 and (
            self.precondition_schedule["time_ends"]
            - self.precondition_schedule["time_start"]
        ).max() > 1800:
            print("Precondition duration is higher than configured")
            valid = False

        # Check precondition energy
        precon_energy = (
            self.precondition_schedule["time_ends"]
            - self.precondition_schedule["time_start"]
        ) * self.precondition_schedule["charging_power"]
        precon_energy.sum()

        # Check energy_real (energy in charging processes) bigger than sum of energy sorted in intervalls
        if total_energy_real > results.at["Energy (kWh)", "Smart Charging"]:
            print(
                "Violation of energy conservation rate. The buses loaded with more enegry then the grid was providing. Smart Charging"
            )
            valid = False

        # Check energy_real (energy in charging processes) bigger than sum of energy sorted in intervalls
        if total_energy_real > results.at["Energy (kWh)", "Imm. Charging"]:
            print(
                "Violation of energy conservation rate. The buses loaded with more enegry then the grid was providing. Immediately Charging"
            )
            valid = False

        if (
            1
            - abs(
                results.at["Energy (kWh)", "Imm. Charging"]
                / results.at["Energy (kWh)", "Smart Charging"]
            )
            > 0.01
        ):
            print("Immediately Charging and Smart Charging differs more then 1 percent")
            valid = False

        for vID, df in self.charging_log.items():
            if df["power"].nunique() >= 8:
                print(
                    "Bus "
                    + vID
                    + " charged with "
                    + str(df["power"].nunique())
                    + " different power values."
                )
                valid = False

        print("Smart charging is valid: " + str(valid))


class ControlSmartCharging:
    """Some control units for the Smart Charging class. Does multiple runs with the same basic settings."""

    def __init__(
        self,
        data,
        start_date,
        price_data_path,
        power_limit_grid=0,
        charging_efficiency=1,
    ):
        """

        :param simulation_host:
        :param start_date:
        :param price_data_path:
        :param power_limit_grid: int or list of powerlimits (15 min intervalls) [or an function] not yet implemnted
        :param capacity_charge: "Jahresleistungspreis" in EUR/kW*a
        :param charging_efficiency:
        """
        self.smart_charging = eflips.depot.SmartCharging(
            data, start_date, price_data_path, power_limit_grid, charging_efficiency
        )
        self.power_limit_gird = power_limit_grid

    def lowest_power(self, accuracy):
        """
        Uses as the first upper boundary the max power of immediately charging, if no power_limit_grid is given, only works if the power_limit_grid is an int
        :param accuracy: in percent
        :return: results of the last succesful power run, power_lower_boundary and power_upper_boundary
        """
        accuracy = accuracy / 100
        if not (isinstance(self.power_limit_grid, int)):
            raise Exception(
                "Function (lowest_power) only works if, the power_limit_grid is an integer."
            )
        if self.power_limit_gird <= 0:
            potential_power_boundary = self.power.max().used_power_immediately
        else:
            potential_power_boundary = self.power_limit_gird
        power_upper_boundary = potential_power_boundary
        power_lower_boundary = 1  # because of not dividing by zero

        self.smart_charging.set_power_limit_grid(potential_power_boundary)
        if self.smart_charging.smart_charging_algorithm():
            power_upper_boundary = potential_power_boundary
            potential_power_boundary = (
                power_upper_boundary - (power_upper_boundary - power_lower_boundary) / 2
            )
            deviation = (
                power_upper_boundary - power_lower_boundary
            ) / power_lower_boundary
        else:
            print("Power limit grid is to low.")
            return

        while accuracy < deviation:
            self.smart_charging.set_power_limit_grid(potential_power_boundary)
            if self.smart_charging.smart_charging_algorithm():
                power_upper_boundary = potential_power_boundary
                potential_power_boundary = (
                    power_upper_boundary
                    - (power_upper_boundary - power_lower_boundary) / 2
                )
                deviation = (
                    power_upper_boundary - power_lower_boundary
                ) / power_lower_boundary
            else:
                power_lower_boundary = potential_power_boundary
                potential_power_boundary = (
                    power_upper_boundary
                    - (power_upper_boundary - power_lower_boundary) / 2
                )
                deviation = (
                    power_upper_boundary - power_lower_boundary
                ) / power_lower_boundary

        # Just for safety.
        self.smart_charging.set_power_limit_grid(power_upper_boundary)
        self.smart_charging.smart_charging_algorithm()

        self.smart_charging.immediately_charging()

        self.smart_charging.smart_charging_algorithm_even()
        self.smart_charging.plot_results()
        self.smart_charging.plot_charging_intervalls_smart()
        self.smart_charging.plot_charging_intervalls_smart_even()

        return self.smart_charging.results(), power_lower_boundary, power_upper_boundary



# -*- coding: utf-8 -*-
"""
Created on Sun Oct  1 12:50:18 2017

@author: p.mundt, e.lauth

Complementary components that are necessary for a standalone run of the depot
simulation.

"""
from math import ceil
from warnings import warn

from eflips.evaluation import DataLogger
from eflips.helperFunctions import flexprint
from eflips.settings import globalConstants
from xlrd import open_workbook

from eflips.depot.depot import VehicleFilter, BackgroundStore
from eflips.depot.simple_vehicle import SimpleVehicle


class VehicleGenerator(BackgroundStore):
    """Initialize SimpleVehicle objects at the start of the simulation.

    Attributes:
    items: [list] containing all generated vehicles

    """

    def __init__(self, env):
        super(VehicleGenerator, self).__init__(env, "VehicleGenerator")
        self.vIDCounter = {}
        self.map_depots = None

    def _complete(self, depots):
        """Preparations that must take place before self.run and simulation
        start, but may not be possible during the depot configuration phase.
        """
        self.map_depots = {depot.ID: depot for depot in depots}

        if globalConstants["general"]["LOG_ATTRIBUTES"]:
            self.logger = DataLogger(self.env, self, "BACKGROUNDSTORE")

    def run(self, depots):
        """Intialize vehicles based on data in eflips settings. Executed and
        completed at simulation start.

        depots: [list] of Depot objects
        """
        self._complete(depots)

        vehicle_count = globalConstants["depot"]["vehicle_count"]

        for depotID in vehicle_count:
            if depotID in self.map_depots:
                home_depot = self.map_depots[depotID]

                for vtID in vehicle_count[depotID]:
                    if vtID not in self.vIDCounter:
                        self.vIDCounter[vtID] = 0

                    for no in range(vehicle_count[depotID][vtID]):
                        self.vIDCounter[vtID] += 1

                        # Initialize vehicle object
                        vehicle = SimpleVehicle(
                            self.env,
                            vtID + " " + str(self.vIDCounter[vtID]),
                            next(
                                vt
                                for vt in globalConstants["depot"]["vehicle_types_obj"]
                                if vt.ID == vtID
                            ),
                            home_depot,
                        )

                        # Assign vehicle to the depot
                        self.put(vehicle)
                        home_depot.init_store.put(vehicle)

            else:
                warn(
                    "HomeDepotID '%s' did not match any ID of known depots." % depotID
                    + " Vehicle entry skipped."
                )

        self.check_arrival()

        # Calculate count and share of vehicles for vehicle types and vehicle
        # type groups
        for depotID in vehicle_count:
            depot = self.map_depots[depotID]
            total = sum(vehicle_count[depotID].values())

            for vtID, count in vehicle_count[depotID].items():
                vehicle_type = next(
                    vt
                    for vt in globalConstants["depot"]["vehicle_types_obj"]
                    if vt.ID == vtID
                )

                vehicle_type.count[depot] = count
                vehicle_type.share[depot] = count / total

        for vtg in globalConstants["depot"]["vehicle_type_groups"]:
            for vehicle_type in vtg.types:
                for depot in vehicle_type.count:
                    if depot not in vtg.count:
                        vtg.count[depot] = 0

                    vtg.count[depot] += vehicle_type.count[depot]

            for depot in vtg.count:
                total = sum(vehicle_type.count[depot] for vehicle_type in vtg.types)
                vtg.share[depot] = total / sum(vehicle_count[depot.ID].values())

    def check_arrival(self):
        for depot in self.map_depots.values():
            if depot.init_store.count > depot.default_plan[0].capacity:
                warn(
                    "The first default plan entry '%s' of depot '%s' should "
                    "have a capacity at least as high as the amount of "
                    "vehicles used in the depot." % (depot.default_plan[0].ID, depot.ID)
                )


class SimpleTrip:
    """Data for one trip outside the depot, from departure until arrival.

    Parameters:
    origin, destination: Depot ID [str] upon init. Is converted to [Depot]
        instance by Timetable
    vehicle_types: [list] of SimpleVehicle.VehicleType as [str] upon import,
        converted to [VehicleType] before simulation start.
    std: [int] scheduled time of departure in total seconds since sim start
    sta: [int] scheduled time of arrival in total seconds since sim start
    distance: [int] total trip distance in km
    start_soc, end_soc: [int or float or None] soc upon departure from and
        arrival at the depot. Must be provided if globalConstants['depot'][
        'consumption_calc_mode'] == 'soc_given'. May be omitted if consumption
        calculation is CR-based.
    charge_on_track: [bool] For the use with globalConstants['depot'][
        'consumption_calc_mode'] == 'soc_given'. If True, opportunity charging
        on track is assumed and start_soc is interpreted as sufficient soc for
        departure and end_soc is interpreted as fixed value upon arrival. If
        False, depot charging is assumed and start_soc and end_soc are used to
        determine the required energy. In this case, start_soc must be higher
        than end_soc. May be omitted if consumption calculation is CR-based.

    Attributes:
    eta: [int] scheduled time of arrival
    atd: [int or None] actual time of departure
    ata: [int or None] actual time of arrival
    vehicle: [SimpleVehicle] assigned to before and after departure
    reserved_for_init: see option prioritize_init_store in globalConstants
    vehicle_from: [str] ID of the area the executing vehicle was located at
        when calling checkout.
    is_copy: boolean flag to indicate if this trip is a copy of another trip
    t_match: [None or int] time of matching this trip and a vehicle. If a
        vehicle from a depot's init store is assigned, this attribute stays
        None. If the match changes, only the last matching time is saved.
    got_early_vehicle: [bool] True if the trip was served by a vehicle where
        the battery wasn't fully charged upon departure.
    periodic_trigger_scheduled: [bool] flag to prevent recursion when
        scheduling a dispatch trigger for delayed trips
    ID_orig: [str] ID of trip this trip is a copy from, if not an original

    """

    delay_event_cls = None
    """Event that succeeds if departure delay occurs. Must be set before 
    calling init to be applied."""

    def __init__(
        self,
        env,
        ID,
        line_name,
        origin,
        destination,
        vehicle_types,
        std,
        sta,
        distance,
        start_soc=None,
        end_soc=None,
        charge_on_track=False,
        is_copy=False,
    ):
        self.env = env
        self.ID = ID
        self.line_name = line_name
        self.origin = origin
        self.destination = destination
        self.vehicle_types = vehicle_types
        self.std = std
        self.sta = sta
        self.distance = distance
        self.start_soc = start_soc
        self.end_soc = end_soc
        self.charge_on_track = charge_on_track
        self.is_copy = is_copy

        if start_soc is not None and not charge_on_track and start_soc < end_soc:
            raise ValueError(
                "For depot chargers start_soc cannot be lower "
                "than end_soc. Trip: %s" % ID
            )

        self.eta = sta
        self.atd = None
        self.ata = None

        self.vehicle = None
        self.reserved_for_init = False
        self.vehicle_from = None

        self.t_match = None
        self.got_early_vehicle = False
        self.t_got_early_vehicle = None
        self.periodic_trigger_scheduled = False

        if self.delay_event_cls is not None:
            self.env.process(self.notify_delay())

    def __repr__(self):
        return "{%s} %s" % (type(self).__name__, self.ID)

    @property
    def vehicle_types_str(self):
        """Return IDs of self.vehicle_types in a list."""
        return [vt.ID for vt in self.vehicle_types]

    @property
    def vehicle_types_joinedstr(self):
        """Return the IDs of self.vehicle_types comma-separated in a single
        string.
        Example returns:
        ['EN'] -> 'EN'
        ['EN', 'DL'] -> 'EN, DL'
        """
        return ", ".join(vt.ID for vt in self.vehicle_types)

    @property
    def delay_departure(self):
        """Return the departure delay [int] in seconds. Return None until
        scheduled time of departure.
        """
        if self.env.now < self.std:
            return None
        else:
            if self.atd is None:
                # trip hasn't started yet and is due or delayed
                return self.env.now - self.std
            else:
                # trip has started
                return self.atd - self.std

    @property
    def delayed_departure(self):
        """Return True if the trip is delayed upon departure or False if not.
        Return None until scheduled time of departure.
        """
        delay = self.delay_departure
        if delay is None:
            return None
        else:
            return delay > 0

    @property
    def delay_arrival(self):
        """Return the arrival delay [int] in seconds. Return None until
        scheduled time of arrival.
        """
        if self.env.now < self.sta:
            return None
        else:
            if self.ata is None:
                return self.env.now - self.sta
            else:
                return self.ata - self.sta

    @property
    def delayed_arrival(self):
        """Return True if the trip is delayed upon arrival or False if not.
        Return None until scheduled time of arrival.
        """
        delay = self.delay_arrival
        if delay is None:
            return None
        else:
            return delay > 0

    @property
    def duration(self):
        """Return the scheduled duration [int] of the trip."""
        return self.sta - self.std

    @property
    def actual_duration(self):
        """Return the actual duration [int] of the trip. Return None until the
        trip is finished."""
        return self.ata - self.atd if self.ata is not None else None

    @property
    def lead_time_match(self):
        """Return the time interval [int] of how early the trip was matched
        with a vehicle. May be negative if the trip is delayed. Return None if
        self.t_match is None.
        """
        if self.t_match is None:
            return None
        else:
            return self.std - self.t_match

    def notify_due_departure(self, t):
        """Wait until simulation time *t* and then trigger the depot's
        assignment process if the trip has no vehicle. No action if *t* < now.
        """
        if t >= self.env.now:
            yield self.env.timeout(t - self.env.now)
            if self.vehicle is None:
                self.origin.depot_control.trigger_dispatch()

    def notify_delay(self):
        """Instantiate self.delay_event_cls if the trip hasn't started one
        second after self.std.
        """
        yield self.env.timeout(self.std + 1 - self.env.now)
        if self.atd is None:
            delay_event = self.delay_event_cls(self.env, self)
            delay_event.trip = self
            delay_event.succeed()


class Timetable:
    """Timetable with a list of SimpleTrip objects from imported data.
    Processes the list of trips and requests matching vehicles from a depot.

    Parameters:
        trips: [list] of SimpleTrip objects. Has to be sorted by departure time
        in ascending order. Only contains trips that are not copies.
    days_ahead: [int] number of days that trips will be issued early. For
        making depot-side planning possible. Should be higher than
        *interval_covered* (in days).

    Attributes:
    trips_issued: [list] of all trips issued to depots. Is empty at the start
        and filled during simulation.
    repetitions: [int] number of times the basic list of trips was
        copied during runtime
    interval_covered: [int] days in seconds that are covered by trips
    all_trips: working [list] of all trips after scheduling including copies
    reservations: {dict} helper variable for the prioritize_init_store option
    fully_reserved: [bool] helper variable for the prioritize_init_store option

    """

    def __init__(self, env, trips):
        self.env = env
        self.trips = trips
        self.repetitions = 0

        interval_covered_sharp = self.trips[-1].std - self.trips[0].std
        self.interval_covered = int(86400 * ceil(interval_covered_sharp / 86400))
        self.trips_issued = []

        self.all_trips = self.trips.copy()
        self.reservations = {}
        self.fully_reserved = False

    def _complete(self, depots):
        """Completion of trip instantiation that must take place before
        simulation start, but may not be possible during the depot
        configuration phase.
        """
        # Convert depot IDs of trips from str to actual Depot objects
        map_ids = {depot.ID: depot for depot in depots}
        for trip in self.trips:
            trip.origin = map_ids[trip.origin]
            trip.destination = map_ids[trip.destination]

        # Check if VehicleGenerator.complete has been called
        if "vehicle_types_obj" not in globalConstants["depot"]:
            raise ValueError(
                "VehicleGenerator.complete must be called before " "Timetable.complete."
            )

        # Check if trip data and vehicle data are valid
        for trip in self.trips:
            for vtID in trip.vehicle_types:
                if vtID not in globalConstants["depot"]["vehicle_types"]:
                    raise ValueError(
                        'vehicle_types "%s" of trip %s is not listed in'
                        % (vtID, trip.ID)
                        + " globalConstants['depot']['vehicle_types']"
                    )

        # Convert trip.vehicle_types from a list of str to a list of
        # VehicleType objects
        for trip in self.trips:
            trip.vehicle_types = [
                next(
                    vt
                    for vt in globalConstants["depot"]["vehicle_types_obj"]
                    if vt.ID == vtID
                )
                for vtID in trip.vehicle_types
            ]

        # Preparation for prioritize_init_store
        if globalConstants["depot"]["prioritize_init_store"]:
            self.reservations = {
                depotID: {
                    vtID: 0
                    for vtID in globalConstants["depot"]["vehicle_count"][depotID]
                }
                for depotID in globalConstants["depot"]["vehicle_count"]
            }
            self.reserve_trips(self.trips)

    def run(self, depots):
        """Assure that trips are issued at depots with a time buffer of
        self.days_ahead.
        Infinite loop that checks if there is enough buffer. If not, copy the
        basic list of trips, raise the time values and issue them at the
        depots.
        *depots*: [list] ob Depot objects. Must contain all depots that are
        mentioned as trip attribute.
        """
        self._complete(depots)
        self.issue_requests(self.trips)

        while True:
            yield self.env.timeout(self.interval_covered)

    def issue_requests(self, trips):
        """Issue requests for all trips in trips."""
        for trip in trips:
            self.issue_request(trip)
            self.trips_issued.append(trip)

    @staticmethod
    def issue_request(trip):
        """Issue request for vehicle at the trip's departure depot. The vehicle
        has to match *filter*. Waiting time is possible. Called in run().
        """
        # Please dont add restrictions here!
        # Go to dispatch strategies in depot.py instead.
        vf = VehicleFilter(filter_names=["trip_vehicle_match"], trip=trip)

        trip.origin.request_vehicle(trip, filter=vf)

    def reserve_trips(self, trips):
        """Mark the first trips equal to the amount of matching vehicles to
        be served by the depot's init store.
        Call only if option prioritize_init_store is on.
        """
        if not self.fully_reserved:
            for trip in trips:
                # look for the first best match
                for vt in trip.vehicle_types:
                    if (
                        self.reservations[trip.origin.ID][vt.ID]
                        < globalConstants["depot"]["vehicle_count"][trip.origin.ID][
                            vt.ID
                        ]
                    ):
                        self.reservations[trip.origin.ID][vt.ID] += 1
                        trip.reserved_for_init = True
                        # vehicle found, stop search for this trip
                        break
            flexprint(
                "reservations: %s" % self.reservations, env=self.env, switch="dispatch2"
            )

            # Check if there are any available vehicles left to possibly
            # shorten next method call (checking if init store is empty is not
            # safe if interval_covered is very small)
            for depotID in self.reservations:
                for vtID in self.reservations[depotID]:
                    if (
                        self.reservations[depotID][vtID]
                        < globalConstants["depot"]["vehicle_count"][depotID][vtID]
                    ):
                        break
                else:
                    continue
                # break outer loop if break was called in the inner loop
                break
            else:
                self.fully_reserved = True


class ExcelSheetData:
    """Methods to import an Excel sheet and save its data row-wise to a list.

    data: [list] containing the imported data with one entry per row.

    """

    def __init__(self, filename, sheetname):
        self.filename = filename
        self.sheetname = sheetname
        self.data = []

        self.import_data()

    def import_data(self):
        wb = open_workbook(self.filename)
        sheet = wb.sheet_by_name(self.sheetname)

        for row_no in range(sheet.nrows):
            row = sheet.row(row_no)
            row_values = []
            for col_no in range(len(row)):
                row_values.append(row[col_no].value)
            self.data.append(row_values)

    def check_for_same_length(self):
        """Return True if all rows have the same length, else False."""
        length = len(self.data[0])
        for row in self.data[1:]:
            if len(row) != length:
                return False
        return True

    def map_headers(self):
        headers = self.data[0]
        datamap = {}
        for no, header in enumerate(headers):
            datamap[header] = no
        return datamap


def timetabledata_from_excel(filename):
    """Load data from excel to use for timetable initialization. The loaded
    data can be reused for multiple identical independent timetables.
    Return an ExcelSheetData object.

    filename: [str] name of excel file including path, excluding file
    extension. Example:
        '..\\bvg_depot\\tripdata\\tripData_depot_I_vehicleType_EN_GN_DL'
    """
    filename += ".xlsx"
    return ExcelSheetData(filename, "Tripdata")


def timetable_from_timetabledata(env, timetabledata):
    """Initialize and return a Timetable instance with SimpleTrip objects
    based on *timetabledata*.

    timetabledata: [ExcelSheetData] from timetabledata_from_excel()
    """
    trips = timetabledata_to_trips(env, timetabledata)
    timetable = Timetable(env, trips)
    return timetable


def timetabledata_to_trips(env, timetabledata):
    """Convert route data from excel to a list of SimpleTrip objects usable by
    Timetable.

    timetabledata: [ExcelSheetData] object that contains a list of trips. Has
        to be sorted by std in ascending order because run() relies on
        this order. In init(), trips is created from this data.
    """
    assert timetabledata.check_for_same_length()
    datamap = timetabledata.map_headers()

    # Instantiate SimpleTrip objects from the Excel data
    trips = []
    for entry in timetabledata.data[1:]:
        # At least one of trip's vehicle_types entry must be specified in
        # globalConstants
        if any(
            [
                vt in globalConstants["depot"]["vehicle_types"].keys()
                for vt in entry[datamap["vehicle_types"]].split(", ")
            ]
        ):
            trips.append(
                SimpleTrip(
                    env,
                    str(entry[datamap["ID"]]),
                    str(entry[datamap["line_name"]]),
                    entry[datamap["origin"]],
                    entry[datamap["destination"]],
                    entry[datamap["vehicle_types"]].split(", "),
                    int(round(entry[datamap["std [s]"]])),
                    int(round(entry[datamap["sta [s]"]])),
                    entry[datamap["distance [km]"]],
                    entry[datamap["start_soc"]] if "start_soc" in datamap else None,
                    entry[datamap["end_soc"]] if "end_soc" in datamap else None,
                    entry[datamap["charge_on_track"]]
                    if "charge_on_track" in datamap
                    else False,
                )
            )
        else:
            raise ValueError(
                "No matching vehicle type for trip %s (filename=%s) is "
                "defined in globalConstants (filename=%s)."
                % (
                    entry[datamap["ID"]],
                    timetabledata.filename,
                    globalConstants["FILENAME_SETTINGS"],
                )
            )
    return trips
